import logging
import toml
import sys
from pathlib import Path
from toml import TomlDecodeError
import os
from monstr.encrypt import Keys
from monstr.util import ConfigError
from monstr.ident.keystore import SQLiteKeyStore, KeyDataEncrypter, KeystoreInterface
from monstr.signing.signing import SignerInterface, BasicKeySigner
from monstr.signing.nip46 import NIP46Signer
from getpass import getpass


def load_toml(filename, dir, current_args):
    if os.path.sep not in filename:
        filename = dir+os.path.sep+filename

    f = Path(filename)
    if f.is_file():
        try:
            toml_dict = toml.load(filename)

            for n, v in toml_dict.items():
                n = n.replace('-','_')
                if n in current_args:
                    c_v = current_args[n]
                    if isinstance(c_v, dict):
                        for k, v in v.items():
                            c_v[k] = v
                    else:
                        current_args[n] = v

                else:
                    current_args[n] = v

        except TomlDecodeError as te:
            print(f'Error in config file {filename} -{te}')
            sys.exit(2)

    else:
        logging.debug(f'load_toml:: no config file {filename}')


async def get_key_from_str(key: str,
                           private_only=False,
                           key_store: KeystoreInterface = None) -> Keys:
    """
    given key str try to make a Key obj
    :param key:            , seperated nsec/npub
    :param private_only:    only accept keys that we can sign with
    :param key_store:
    :return: Key
    """

    # if it's a bech32 str then convert to key obj
    if Keys.is_bech32_key(key):
        ret = Keys.get_key(key)
    # else we'll have a look in the key_store if one was given
    elif key_store:
        ret = await key_store.get(key)
        if ret is None:
            raise ConfigError(f'{key} doesn\'t look like a nsec/npub nostr key or alias not found')
    else:
        raise ConfigError(f'{key} doesn\'t look like a nsec/npub nostr key')

    if private_only and ret.private_key_hex() is None:
        raise ConfigError(f'{key} is not a private key')

    return ret


async def get_signers_from_str(keys: str, key_store: KeystoreInterface = None) -> [SignerInterface]:
    if ',' in keys:
        key_arr = keys.split(',')
    else:
        key_arr = [keys]

    ret = []
    for c_key_str in key_arr:
        if c_key_str.startswith('bunker://'):
            ret.append(NIP46Signer(c_key_str, auto_start=True))
        else:
            ret.append(BasicKeySigner(await get_key_from_str(key=c_key_str,
                                                             key_store=key_store,
                                                             private_only=True)))

    return ret


async def get_keys_from_str(keys: str,
                            private_only=False,
                            key_store: KeystoreInterface = None) -> [Keys]:
    """
    get Key objects from string e.g. that can be npub,nsec or alias and may be mutiple seperated by comma
    :param key_store:
    :param keys:            , seperated nsec/npub
    :param private_only:    only accept nsec
    :param single_only:     only a single key
    :return: [Key]
    """

    if ',' in keys:
        key_arr = keys.split(',')
    else:
        key_arr = [keys]
    return [await get_key_from_str(key=k,
                                   private_only=private_only,
                                   key_store=key_store)for k in key_arr]


def get_sqlite_key_store(db_file, password: str = None):
    # human alias to keys
    # keystore for user key aliases
    async def get_key() -> str:
        ret = password
        if password is None:
            ret =  getpass('keystore key: ')
        return ret

    key_enc = KeyDataEncrypter(get_key=get_key)
    return SQLiteKeyStore(file_name=db_file,
                          encrypter=key_enc)



"""
watches one relay and post what it sees into another
"""
import logging
import asyncio
from datetime import datetime,timedelta
from monstr.util import util_funcs
from monstr.client.client import Client, ClientPool
from monstr.client.event_handlers import RepostEventHandler


async def do_mirror(from_url,
                    to_url,
                    base_filter=None,
                    as_users=None,
                    since=datetime.now()):

    def get_filter():
        ret = base_filter
        if as_users:
            # here we should fetch contacts of as_users and add all returned pubks to authors in filter
            pass

        return ret

    async with ClientPool(from_url) as from_relay:
        print('connected to from relay')
        use_filter = get_filter()


        # now start mirroring to to_url
        async with ClientPool(to_url) as to_relay:
            print('mirroring from %s to %s with filter %s' % (from_url,
                                                              to_url,
                                                              use_filter))
            my_repost = RepostEventHandler(to_relay)

            def on_connect(the_from_relay):
                the_from_relay.subscribe('my_mirror',
                                         handlers=my_repost,
                                         filters=use_filter)

            # manually call to make the sub
            on_connect(from_relay)

            while True:
                await asyncio.sleep(0.5)


    # # where we're posting to
    # to_relay = ClientPool(to_relay)
    # asyncio.create_task(to_relay.run())
    #
    # # TODO add EOSE support
    # my_repost = RepostEventHandler(to_relay)
    # def on_connect(the_from_relay):
    #     the_from_relay.subscribe(handlers=my_repost, filters=get_filter())
    #
    # from_relay = ClientPool(from_relay, on_connect=on_connect)
    # await from_relay.run()
    #
    # print('starting mirror \nfrom %s to %s \nwith filter=%s' % (from_relay, to_relay, filter))

if __name__ == "__main__":
    logging.getLogger().setLevel(logging.DEBUG)

    # from_relay = ['ws://localhost:8082/','ws://localhost:8083/']
    from_relay = ['wss://nos.lol']
    to_relay = ['ws://localhost:8081/']
    # events will be fetch from this time point, default now
    since = None
    if since is None:
        since = util_funcs.date_as_ticks(datetime.now())

    # base filter for example to only mirror set kinds
    # {kinds: [Event.KIND..., ]}
    base_filter = {
        'since': since
    }

    # if given the kind 3 for this keys will be looked up and
    # the base query will be modified to request from aithors
    # mentioned in the returned contacts, default any
    as_user = None

    asyncio.run(do_mirror(from_relay, to_relay,
                          base_filter=base_filter))



import logging
import asyncio
from pathlib import Path
import argparse
from monstr.ident.profile import Profile
from monstr.ident.event_handlers import NetworkedProfileEventHandler
from monstr.client.client import ClientPool, Client
from monstr.event.event import Event
from monstr_terminal.app.post import PostApp
from monstr.inbox import Inbox
from monstr_terminal.cmd_line.post_loop_app import PostAppGui
from monstr.encrypt import Keys, DecryptionException
from monstr.signing.signing import BasicKeySigner, SignerInterface
from monstr.util import util_funcs, ConfigError
from monstr_terminal.util import load_toml, get_sqlite_key_store
from monstr.ident.keystore import KeystoreInterface
"""
    TODO: add support for NIP44 encryption 
"""

# defaults if not otherwise given
# working directory it'll be created it it doesn't exist
WORK_DIR = f'{Path.home()}/.nostrpy/'
# filename for key store
KEY_STORE_DB_FILE = 'keystore.db'
# relay/s to attach to
RELAYS = 'ws://localhost:8081'
# user to post as
USER = None
# config from toml file
CONFIG_FILE = 'poster.toml'
# default kind for inbox wrap events
INBOX_KIND = Event.KIND_ENCRYPT


async def create_key(key_val: str, for_desc: str, key_store: KeystoreInterface = None) -> Keys:
    ret = Keys.get_key(key_val)
    if ret is None and key_store:
        ret = await key_store.get(key_val)

    if ret is None:
        raise ConfigError(f'unable to create {for_desc} keys using - {key_val}')

    if ret.private_key_hex() is None:
        raise ConfigError(f'unable to create {for_desc} private keys using - {key_val}')
    return ret


async def get_user_signer(user: str,
                          for_desc: str = None,
                          key_store: KeystoreInterface = None) -> SignerInterface:

    # special ?, we'll just create some random keys to post as
    if user == '?':
        adhoc_k = Keys()
        ret = BasicKeySigner(adhoc_k)
        print(f'created adhoc key for {for_desc} - {adhoc_k.private_key_bech32()}')
    else:
        ret = BasicKeySigner(await create_key(key_val=user,
                                              for_desc=for_desc,
                                              key_store=key_store))

    return ret


async def get_to_keys(to_users: [str], ignore_missing: bool, key_store: KeystoreInterface = None) -> [Keys]:
    ret = []
    if to_users is not None:
        for c_to in to_users:
            # get rid of anyspaces
            # TODO: keystore needs to enforce non space in key names
            c_to = c_to.replace(' ', '')

            cu_key = Keys.get_key(c_to)
            if cu_key is None and key_store:
                cu_key = await key_store.get(c_to)

            if cu_key is None:
                # just note
                if ignore_missing:
                    logging.info(f'unable to create keys for to_user using - {c_to} - ignoring')
                # bug out
                else:
                    raise ConfigError(f'unable to create keys for to_user using - {c_to}')
            else:
                ret.append(cu_key)

        # even with ignore_missing we need atleast 1 to user
        if not ret:
            raise ConfigError('unable to create any to user keys!')

    return ret


async def get_poster(client: Client,
                     profile_handler: NetworkedProfileEventHandler,
                     user_sign: SignerInterface,
                     to_users_k: [Keys],
                     inbox_sign: SignerInterface,
                     inbox_kind: int,
                     is_encrypt: bool,
                     subject: str,
                     kind: int,
                     tags: str) -> PostApp:
    k: Keys

    # are we going via an inbox?
    inbox = None
    if inbox_sign:
        inbox = await get_inbox(inbox_sign=inbox_sign,
                                user_sign=user_sign,
                                to_keys=to_users_k,
                                inbox_kind=inbox_kind,
                                profile_handler=profile_handler)

    post_app = PostApp(use_relay=client,
                       profile_handler=profile_handler,
                       user_sign=user_sign,
                       to_users_k=to_users_k,
                       inbox=inbox,
                       subject=subject,
                       is_encrypt=is_encrypt,
                       kind=kind,
                       tags=tags)

    await post_app.wait_ready()

    return post_app


async def get_inbox(inbox_sign: SignerInterface,
                    user_sign: SignerInterface,
                    to_keys: [Keys],
                    inbox_kind: int,
                    profile_handler: NetworkedProfileEventHandler=None) -> Inbox:
    # are we going via an inbox?
    ret = None
    inbox_pub_k = await inbox_sign.get_public_key()

    # if given a profile handler we'll attempt to get a name for the inbox (from meta event)
    if profile_handler is not None:
        inbox_p: Profile = await profile_handler.aget_profile(inbox_pub_k, create_missing=True)
        name = inbox_p.display_name()

    # actually create the inbox
    ret = Inbox(signer=inbox_sign,
                name=name,
                use_kind=inbox_kind)

    # generate share map for any encrypts over the inbox
    if to_keys:
        await ret.set_share_map(for_sign=user_sign,
                                to_keys=to_keys)

    return ret


async def post_single(relays: [str],
                      user_sign: SignerInterface,
                      to_users_k: Keys,
                      inbox_sign: SignerInterface,
                      inbox_kind: int,
                      is_encrypt: bool,
                      subject: str,
                      message: str,
                      kind: int = None,
                      tags: str = None):

    def on_auth(the_client: Client, challenge: str):
        auth_sign = user_sign
        if inbox_sign:
            auth_sign = inbox_sign

        asyncio.create_task(the_client.auth(auth_sign, challenge))

    async with ClientPool(relays, timeout=10, on_auth=on_auth) as client:
        try:
            post_app = await get_poster(client=client,
                                        profile_handler=NetworkedProfileEventHandler(client=client),
                                        user_sign=user_sign,
                                        to_users_k=to_users_k,
                                        inbox_sign=inbox_sign,
                                        inbox_kind=inbox_kind,
                                        is_encrypt=is_encrypt,
                                        subject=subject,
                                        kind=kind,
                                        tags=tags)

            await post_app.show_post_info(message)
            await post_app.do_post(message)
        except Exception as e:
            print(e)



        # hack to give time for the event to be sent
        await asyncio.sleep(1)


async def post_loop(relays: [str],
                    user_sign: SignerInterface,
                    to_users_k: [Keys],
                    inbox_sign: SignerInterface,
                    inbox_kind: int,
                    is_encrypt: bool,
                    subject: str,
                    kind: int,
                    tags: str):

    sub_id: str = None
    con_status = None

    inbox_pub_k = None
    if inbox_sign:
        inbox_pub_k = await inbox_sign.get_public_key()

    # if we're using an inbox then events are always encrypted type 4
    # though they may be unencrypted to anyone who has the inbox keys
    def do_sub():
        nonlocal sub_id

        filter = {
            'kinds': [kind],
            'limit': 10
        }
        if inbox_pub_k:
            filter['authors'] = [inbox_pub_k]
            filter['kinds'] = [inbox_kind]

        # can only be applied on non wrapped, otherwise needs to be filtered by post app
        if subject and inbox_pub_k is None:
            filter['#subject'] = subject

        sub_id = my_client.subscribe(handlers=[post_app],
                                     filters=filter,
                                     sub_id=sub_id)

    def on_status(status):
        nonlocal con_status
        if con_status != status['connected']:
            con_status = status['connected']
            asyncio.create_task(my_gui.draw_messages())

    def on_eose(the_client: Client, sub_id:str, evts: [Event]):
        Event.sort(evts=evts,
                   inplace=True,
                   reverse=False)
        c_evt: Event
        u_authors = list({c_evt.pub_key for c_evt in evts})
        # batch get authors otherwise requests would be fire 1 by 1 as needed
        # and the relay is likely to error us on number of subs
        async def do_events():
            await peh.aget_profiles(u_authors,
                                    create_missing=True)
            for c_evt in evts:
                post_app.do_event(the_client, sub_id, c_evt)

        asyncio.create_task(do_events())

    def on_connect(the_client: Client):
        do_sub()

    def on_auth(the_client: Client, challenge: str):
        auth_sign = user_sign
        if inbox_sign:
            auth_sign = inbox_sign

        asyncio.create_task(the_client.auth(auth_sign, challenge))


    # EOSE based fetch of historic events doesn't work very well with multiple clients
    # TODO: change to initial query then sub based?
    async with ClientPool(relays,
                          on_auth=on_auth,
                          on_eose=on_eose) as my_client:
        peh = NetworkedProfileEventHandler(client=my_client)
        post_app = await get_poster(client=my_client,
                                    profile_handler=peh,
                                    user_sign=user_sign,
                                    to_users_k=to_users_k,
                                    inbox_sign=inbox_sign,
                                    inbox_kind=inbox_kind,
                                    is_encrypt=is_encrypt,
                                    subject=subject,
                                    kind=kind,
                                    tags=tags)

        my_gui = PostAppGui(post_app,
                            profile_handler=peh,
                            is_encrypt=is_encrypt)

        my_client.set_on_status(on_status)
        # my_client.set_on_eose(on_eose)
        my_client.set_on_connect(on_connect)

        # manually call the connect which just adds the sub
        on_connect(my_client)
        await my_gui.run()


def get_cmdline_args(args) -> dict:
    parser = argparse.ArgumentParser(
        prog='post.py',
        description="""
            post nostr events from the command line
            """
    )
    parser.add_argument('-r', '--relay', action='store', default=args['relay'],
                        help=f'comma separated nostr relays to connect to, default [{args["relay"]}]')
    parser.add_argument('-u', '--user', action='store', default=args['user'],
                        help=f"""
                        alias, priv_k of user to post as,
                        default [{args['user']}]""")
    parser.add_argument('-t', '--to_users', action='store', default=args['to_users'],
                        help=f"""
                        comma seperated alias, priv_k, or pub_k of user to post to,
                        default [{args['to_users']}]""")
    parser.add_argument('-v', '--via', action='store', default=args['via'],
                        help=f"""
                            alias(with priv_k) or nsec that will be used as public inbox with wrapped events,
                            default [{args['via']}]""")
    parser.add_argument('-s', '--subject', action='store', default=args['subject'],
                        help=f"""
                                add subject tag to post,,
                                default[{args['subject']}]""")
    parser.add_argument('--tags', action='store', default=args['tags'],
                        help=f"""
                                tags to add post in format tagname:v1,v2#tagname:v1...
                                default [{args["tags"]}]
    """)
    parser.add_argument('message', type=str, nargs='*',
                       help='message to post')
    parser.add_argument('-k', '--kind', action='store', help="""
                                kind of event to post, if not given used kind depends on format -
                                if default or plaintext then [1] if encrypt then [4]"""
                        , type=int,
                        dest='kind')
    parser.add_argument('--inbox-kind', action='store',
                        help=f'if using an inbox, what kind is used for the wrapping event default [{args["inbox_kind"]}]',
                        type=int,default=int(args["inbox_kind"]),
                        dest='inbox_kind')

    parser.add_argument('-f', '--format', action='store', help="""
    format of the event content if default is selected then events of kind 4 will be encrypted and all other kinds will 
    be plaintext
    """, choices={
        'plaintext',
        'encrypt',
        'default'
    }, default='default')

    parser.add_argument('-i', '--ignore_missing', action='store_true', help='don\'t fail on missing to_users')
    parser.add_argument('-l', '--loop', action='store_true', help='stay open to enter and receive messages')
    parser.add_argument('-d', '--debug', action='store_true', help='enable debug output')

    ret = parser.parse_args()

    return vars(ret)


def get_tags(tags: str) -> [[]]:
    ret = None
    for c_tag in tags.split('#'):
        split_vals = c_tag.split(':')
        if len(split_vals) < 2:
            logging.debug(f'get_tags: ignoring {split_vals} no values')
        else:
            if ret is None:
                ret = []
            ret.append(split_vals)

    logging.debug(f'get_tags: got tag values {ret}')
    return ret


def get_args() -> dict:
    """
    get args to use order is
        default -> toml_file -> cmd_line options

    so command line option is given priority if given

    :return: {}
    """

    ret = {
        'relay': RELAYS,
        'work_dir': WORK_DIR,
        'conf': CONFIG_FILE,
        'user': USER,
        'to_users': None,
        'via': None,
        'format': 'default',
        'kind': None,
        'inbox_kind': INBOX_KIND,
        'ignore_missing': False,
        'subject': None,
        'tags': None,
        'message': None,
        'loop': False,
        'alias_file': KEY_STORE_DB_FILE,
        'keystore': {
            'filename': WORK_DIR + KEY_STORE_DB_FILE,
            'password': None
        },
        'debug': False
    }

    # now form config file if any
    load_toml(filename=ret['conf'],
              dir=ret['work_dir'],
              current_args=ret)

    # now from cmd line
    ret.update(get_cmdline_args(ret))

    # if any tags, see if we can make into anything sensible...
    if ret['tags']:
        ret['tags'] = get_tags(ret['tags'])

    # if not given get a user
    if ret['user'] is None:
        ret['user'] = input('as user: ')

    # work out if we're making encrypted posts
    format = ret['format']
    kind = ret['kind']
    is_encrypt = format == 'encrypt'

    # no kind given, so now we'll choose one, maybe overwrite is_encrypt
    if kind is None:
        if format in {'encrypt', 'default'}:
            kind = 4
            is_encrypt = True
        else:
            kind = 1
    # kind has been given, for format default we'll now override is_encrypt
    else:
        # default which is encrypt only for kind 4
        if format == 'default':
            is_encrypt = kind == 4

    ret['is_encrypt'] = is_encrypt
    ret['kind'] = kind

    # to users are required for encypted msgs, so ask for to users!
    if is_encrypt and ret['to_users'] is None:
        ret['to_users'] = input('to users: ')

    # if we msg we reduce any spacing, maybe we shouldn't bother?
    if ret['message']:
        ret['message'] = ' '.join(ret['message'])

    # msg isn't required in loop mode - in fact it's ignored.. maybe we should still send it if given?
    # anyway if not loop we will need a msg
    elif ret['loop'] is False and not ret['message']:
        ret['message'] = input('message: ')


    # if debug flagged enable now and output args we're running with
    if ret['debug'] is True:
        logging.getLogger().setLevel(logging.DEBUG)
        logging.debug(f'get_args:: running with options - {ret}')

    return ret


async def main(args):
    # post to these relays
    relays = args['relay'].split(',')

    # we'll post as this user
    user = args['user']

    # post to these users
    to_users = args['to_users']
    if to_users:
        to_users = to_users.split(',')

    # msgs wrapped via this in box (priv key that we know and so should those we post to)
    inbox = args['via']
    inbox_kind = args['inbox_kind']

    # don't fail just because we can't find all to_users
    ignore_missing = args['ignore_missing']

    # open gui so we can type mutiple messages, in this case message text isn't required
    loop = args['loop']

    # message text to be sent if not loop mode
    message = args['message']

    # subject text for message/ each message
    subject = args['subject']

    # args we encrypting? (NIP4 at the moment)
    is_encrypt = args['is_encrypt']
    # kind of event to post by default this will be 1 for plaintxt or 4 for encrypt
    kind = args['kind']

    # tags if any to be added to posts
    tags = args['tags']

    # keystore for user key aliases
    key_store = get_sqlite_key_store(db_file=WORK_DIR+args['alias_file'],
                                     password=args['keystore']['password'])

    # user to sign are post events, eventually this might be NIP46 client
    # (we don't have the keys locally)
    user_signer = await get_user_signer(user=user,
                                        for_desc='user',
                                        key_store=key_store)

    # to keys, opt for plain text but required if we're encrypting
    to_keys = await get_to_keys(to_users=to_users,
                                ignore_missing=ignore_missing,
                                key_store=key_store)

    inbox_signer = None
    if inbox:
        inbox_signer = await get_user_signer(user=inbox,
                                             for_desc='inbox',
                                             key_store=key_store)

    if loop is False:
        await post_single(relays=relays,
                          user_sign=user_signer,
                          to_users_k=to_keys,
                          inbox_sign=inbox_signer,
                          inbox_kind=inbox_kind,
                          is_encrypt=is_encrypt,
                          subject=subject,
                          message=message,
                          kind=kind,
                          tags=tags
                          )

    else:
        await post_loop(relays=relays,
                        user_sign=user_signer,
                        to_users_k=to_keys,
                        inbox_sign=inbox_signer,
                        inbox_kind=inbox_kind,
                        is_encrypt=is_encrypt,
                        subject=subject,
                        kind=kind,
                        tags=tags
                        )


if __name__ == "__main__":
    logging.getLogger().setLevel(logging.FATAL)
    util_funcs.create_work_dir(WORK_DIR)
    try:
        asyncio.run(main(get_args()))
    except ConfigError as ce:
        print(ce)
    except DecryptionException as de:
        print(f'bad password or non encrypted store? - {de}')







"""
    prompt_toolkit front end to from app.post import PostApp

"""
import asyncio
from prompt_toolkit import Application
from prompt_toolkit.layout import Layout, ScrollablePane
from prompt_toolkit.layout.containers import HSplit, Window
from prompt_toolkit.layout.controls import BufferControl, FormattedTextControl
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.key_binding import KeyBindings
from monstr.ident.event_handlers import ProfileEventHandlerInterface
from monstr.event.event import Event
from monstr.util import util_funcs
from monstr_terminal.app.post import PostApp
import os


class PostAppGui:

    def __init__(self,
                 post_app: PostApp,
                 profile_handler: ProfileEventHandlerInterface,
                 is_encrypt: bool = True):
        # gui parts
        self._post_app = post_app
        self._profile_handler = profile_handler
        self._app = None
        self._msg_split_con = HSplit([])
        self._msg_scroll_con = ScrollablePane(self._msg_split_con)
        self._make_gui()
        self._post_app.set_on_message(self._on_msg)
        self._msg_display_height = 0
        self._draw_task = None

        # will the events need decrypting?
        self._is_encrypt = is_encrypt

    def _make_gui(self):
        kb = KeyBindings()
        buffer1 = Buffer()

        # self._make_msg_split()
        self._msg_display_height = 0
        self._scroll_bottom()
        root_con = HSplit([
            self._msg_scroll_con,
            Window(content=BufferControl(buffer1), height=3)
        ])
        my_layout = Layout(root_con)

        # ctrl-q to quit, also type exit
        @kb.add('c-q')
        def exit_(event):
            self._app.exit()

        @kb.add('c-up')
        def do_up(e):
            pos = self._msg_scroll_con.vertical_scroll-1
            if pos < 0:
                pos = 0
            self._msg_scroll_con.vertical_scroll = pos

        @kb.add('c-down')
        def do_down(e):
            pos = self._msg_scroll_con.vertical_scroll

            if (self._msg_display_height + 3 - pos) > os.get_terminal_size().lines:
                pos += 1

            self._msg_scroll_con.vertical_scroll = pos

        @kb.add('c-s')
        def post_(event):
            msg = buffer1.text
            if msg.replace(' ', ''):

                if self._post_app.connection_status:
                    asyncio.create_task(self._post_app.do_post(msg))
                else:
                    'in someway make user aware that we dont have a connection to relay...'
                    pass

                buffer1.text = ''

        self._app = Application(full_screen=True,
                                layout=my_layout,
                                key_bindings=kb)

    def _scroll_bottom(self):

        pos = self._msg_scroll_con.vertical_scroll
        d_h = os.get_terminal_size().lines
        if self._msg_display_height + 3 - pos > d_h:
            pos = (self._msg_display_height + 3) - d_h
        self._msg_scroll_con.vertical_scroll = pos

    async def draw_messages(self):
        # has to be asunc because of profile fetching
        await self._make_msg_split()
        self._app.invalidate()
        self._scroll_bottom()

    def _on_msg(self, evt):
        # async def delay_draw():
        #     await asyncio.sleep(0.5)
        #     await self.draw_messages()

        # bit hacky but as the draw might end up doing some fetches this stops too many being fired off
        # this seems to work as is... but the above with an actual delay might work better if this still cause
        # problems
        if self._draw_task:
            self._draw_task.cancel()
            self._draw_task = None
        self._draw_task = asyncio.create_task(self.draw_messages())

    async def _make_msg_split(self):
        """
        make up the components to display the posts on screen
        note that though that though we can only send post of type encrypt/plaintext dependent
        on start options, here the view will show both and user won't can't tell the difference.
        Probably should only show encrypt if in encrypt and vice versa

        :return:
        """
        c_m: Event
        to_add = []

        as_user = self._post_app.as_user
        as_sign = self._post_app.as_signer

        self._msg_display_height = 0
        # prefetch all the profiles we'll need
        if self._profile_handler:
            await self._profile_handler.aget_profiles([c_m.pub_key for c_m in self._post_app.message_events],
                                                      create_missing=True)

        for c_m in self._post_app.message_events:
            content = c_m.content

            prompt_col = 'red'
            content_col = ''
            if c_m.pub_key == as_user.public_key:
                prompt_col = 'cyan'
            if not self._post_app.connection_status:
                content_col = prompt_col = 'gray'

            if self._is_encrypt and as_sign:
                try:
                    use_pub_key = c_m.p_tags[0]

                    # its a message to us
                    if c_m.pub_key != as_user.public_key:
                        use_pub_key = c_m.pub_key

                    try:
                        content = await as_sign.nip4_decrypt(payload=c_m.content,
                                                             for_pub_k=use_pub_key)
                    except Exception as x:
                        content = x

                except Exception as e:
                    # currently in the case of group messages we'd expect this except on those we create
                    # and the 1 msg that was encrypted for us... we can't tell which that is until we try to decrypt
                    content = None

            if content:
                msg_height = len(content.split('\n'))
                msg_profile = None
                if self._profile_handler:
                    msg_profile = await self._profile_handler.aget_profile(c_m.pub_key,
                                                                           create_missing=True)

                prompt_user_text = util_funcs.str_tails(c_m.pub_key, 4)
                if msg_profile:
                    prompt_user_text = msg_profile.display_name()

                to_add.append(
                    HSplit([
                        Window(FormattedTextControl(text=[(prompt_col, '%s@%s' % (prompt_user_text, c_m.created_at))]),
                               height=1),
                        Window(FormattedTextControl(text=[(content_col,content)]), height=msg_height)
                    ])

                )
                self._msg_display_height += msg_height + 1

        self._msg_split_con.children = to_add


    async def run(self):
        await self._app.run_async()




import datetime
import logging
import asyncio
import aioconsole
import sys
import signal
import argparse
from pathlib import Path
from monstr.encrypt import DecryptionException
from monstr_terminal.util import load_toml, get_sqlite_key_store, get_signers_from_str
from monstr.util import ConfigError
from monstr.signing.nip46 import NIP46ServerConnection, NIP46AuthoriseInterface

# work dir, we'll try and create if it doesn't exist
WORK_DIR = f'{Path.home()}/.nostrpy/'
# a toml file to get args from, can be overridden from cmd line
CONFIG_FILE = WORK_DIR + 'signer.toml'
# filename for key store
KEY_STORE_DB_FILE = 'keystore.db'

# from relays and as_user the connect str is created
# e.g. bunker://AS_USER_HEX_PUB?relay=RELAY1&RELAY2 etc
# the other way is to get a connection str from a client TODO when find an example
# relay/s to attach to when bunker style remote connect
RELAYS = 'ws://localhost:8081'
# user to sign as when using remote style connect
USER = None
# print info on each event we get request to sign
VERBOSE = False


def print_auth_info(method:str, params:dict):
    print('method', method)
    print('params', params)


class AuthoriseAll(NIP46AuthoriseInterface):

    def __init__(self, verbose: bool = False):
        self._verbose = verbose

    async def authorise(self, method: str, id: str, params: [str]) -> bool:
        if self._verbose:
            print_auth_info(method, params)
        return True


class BooleanAuthorise(NIP46AuthoriseInterface):

    async def authorise(self, method: str, id: str, params: [str]) -> bool:
        # always verbose
        print_auth_info(method, params)
        accept = await aioconsole.ainput('authorise y/n? ')
        return accept.lower() == 'y'


class TimedAuthorise(BooleanAuthorise):

    def __init__(self, auth_mins = 10, verbose: bool = False):
        self._last_auth_at = None
        self._auth_delta = datetime.timedelta(minutes=auth_mins)
        self._verbose = verbose

    async def authorise(self, method: str, id: str, params: [str]) -> bool:
        if self._verbose:
            print_auth_info(method, params)
        now = datetime.datetime.now()
        ret = True

        # maybe we need to reauth?
        if self._last_auth_at is None or (now - self._last_auth_at) > self._auth_delta:
            ret = await super().authorise(method, id, params)
            if ret:
                self._last_auth_at = now

        return ret


def get_cmdline_args(args) -> dict:
    parser = argparse.ArgumentParser(
        prog='signer.py',
        description="""
            A NIP46 server - signs events on behalf of another client
            """
    )
    parser.add_argument('-c', '--conf', action='store', default=args['conf'],
                        help=f'name com TOML file to use for configuration, default[{args["conf"]}]')
    parser.add_argument('--work-dir', action='store', default=args['work_dir'],
                        help=f'base dir for files used if full path isn\'t given, default[{args["work_dir"]}]')
    parser.add_argument('user', action='store', default=args['user'],
                        nargs='?',
                        help=f"""
                        alias or priv_k that we are signing as
                        default[{args['user']}]""")
    parser.add_argument('-r', '--relay',
                        action='store',
                        default=args['relay'],
                        help=f'comma separated nostr relays to connect to, default[{args["relay"]}]')
    parser.add_argument('-a', '--auth',
                        action='store',
                        default=args['auth'],
                        help=f'action on receiving requests to perform signing operations '
                             f'all - always accept, ask - always ask or int value to ask every n minutes, default[{args["auth"]}]')
    parser.add_argument('-v', '--verbose',
                        action='store_true',
                        help=f'print info on each event that is requested to sign, default[{args["verbose"]}]',
                        default=args['verbose'])
    parser.add_argument('--no-verbose',
                        action='store_false',
                        help='turn off verbose',
                        default=not args['verbose'])
    parser.add_argument('-d', '--debug', action='store_true', help='enable debug output', default=args['debug'])
    ret = parser.parse_args()
    return vars(ret)


def get_args() -> dict:
    """
    get args to use order is
        default -> toml_file -> cmd_line options

    so command line option is given priority if given

    :return: {}
    """

    ret = {
        'work_dir': WORK_DIR,
        'conf': CONFIG_FILE,
        'relay': RELAYS,
        'user': USER,
        'auth': 'all',
        'verbose': VERBOSE,
        'debug': False,
        'keystore': {
            'filename': WORK_DIR + KEY_STORE_DB_FILE,
            'password': None
        }
    }

    # only done to get the work-dir and conf options if set
    ret.update(get_cmdline_args(ret))
    # now form config file if any
    load_toml(filename=ret['conf'],
              dir=ret['work_dir'],
              current_args=ret)
    # # 2pass so that cmdline options override conf file options
    ret.update(get_cmdline_args(ret))

    # if debug flagged enable now
    if ret['debug'] is True:
        logging.getLogger().setLevel(logging.DEBUG)
        logging.debug(f'get_args:: running with options - {ret}')

    if not ret['relay']:
        print('Required argument relay is missing. Use -r or --relay')
        exit(1)

    # force get a user
    if not ret['user'] or ret['user'] == '?':
        ret['user'] = input('as user: ')

    auth = ret['auth'].lower()
    if auth not in ('all', 'ask'):
        try:
            ret['auth'] = int(auth)
        except ValueError:
            raise ConfigError(f'auth most be all, ask or integer value got - {auth}')

    return ret


async def main(args):
    """
        This is set to work where we create the connection string, there is another method where the client that
        wants us to sign creates the connection str and then we connct to that.
        Originally thats what we did and so I'm pretty sure changes to support that method would be pretty minor
        (I think it basically just changes who decides the comm key)
        Anyway this code works as a signer for
            https://nostrudel.ninja/
            sign in -> advanced -> nostr connect -> paste the string there we create

            test with other implementations as find them
    """
    try:
        # store of aliases to keys
        key_store = get_sqlite_key_store(db_file=args['keystore']['filename'],
                                         password=args['keystore']['password'])

        # user we're signing for
        user_sign = (await get_signers_from_str(keys=args['user'],
                                                key_store=key_store))[0]

        # relays to attach to
        relays = args['relay'].split(',')

        # print info about events as we auth
        verbose = args['verbose']

        # create the authoriser if any, this decide how we ask user to proceed on requests for sign ops
        auth_type = args['auth']

        if auth_type == 'ask':
            print('all operations will require manual authorisation')
            my_auth = BooleanAuthorise()
        elif isinstance(auth_type, int):
            my_auth = TimedAuthorise(auth_mins=auth_type, verbose=verbose)
            print(f'operations will require manual authorisation every {auth_type} minutes')
        else:
            my_auth = AuthoriseAll(verbose=verbose)
            print(f'operations will always be authorised')

        my_sign_con = NIP46ServerConnection(signer=user_sign,
                                            relay=relays,
                                            authoriser=my_auth)

        # print out the information needed to connect
        print(f'connect with: {await my_sign_con.bunker_url}')

        def sigint_handler(signal, frame):
            my_sign_con.end()
            sys.exit(0)

        signal.signal(signal.SIGINT, sigint_handler)

        # this is for the other style of connection where the client chooses the comm
        # snort used to have this style, add back in if we can find an example to test against
        # if comm_k is not None:
        #     await my_sign_con.request_connect()

        await my_sign_con.run()

    except ConfigError as ce:
        print(ce)
    except DecryptionException as de:
        print(f'bad password or non encrypted store? - {de}')
    except Exception as e:
        print(e)

if __name__ == "__main__":
    logging.getLogger().setLevel(logging.ERROR)
    try:
        asyncio.run(main(get_args()))
    except ConfigError as ce:
        print(ce)









#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
  Simple json dump of the event viewer demonstrating
  how easy it is to play around with the events read
  from nostr without having to implement a full client.

    Usage when you use a config file:
        $ ./python3 view.py -o json | ./view_consumer.py

    or, if you want to use a specific  nostr relay instance:
        $ python3 view.py -r wss://nos.lol -o json | python3 view_consumer.py
"""

import fileinput
import json
import pprint


if __name__ == '__main__':
    for line in fileinput.input():
        pprint.pprint(json.loads(line))




import logging
import sys
import signal
import asyncio
import argparse
from pathlib import Path
from datetime import datetime, timedelta
from monstr.ident.profile import Profile, Contact, NIP5Helper, NIP5Error
from monstr.ident.event_handlers import NetworkedProfileEventHandler, ProfileEventHandlerInterface
from monstr.ident.profile import ContactList
from monstr.client.client import ClientPool, Client
from monstr.client.event_handlers import DeduplicateAcceptor, LengthAcceptor, \
    NotOnlyNumbersAcceptor, EventHandler, LastEventHandler, FilterAcceptor
from monstr.util import util_funcs, ConfigError
from monstr.inbox import Inbox
from monstr.encrypt import Keys
from monstr.signing.signing import SignerInterface, BasicKeySigner
from monstr.event.event import Event
from monstr_terminal.cmd_line.util import FormattedEventPrinter, JSONPrinter, ContentPrinter
from monstr_terminal.util import load_toml, get_keys_from_str, get_sqlite_key_store

# defaults if not otherwise given
# working directory it'll be created it it doesn't exist
WORK_DIR = f'{Path.home()}/.nostrpy/'
# filename for key store
KEY_STORE_DB_FILE = 'keystore.db'
# relay/s to attach to
RELAYS = None
# user to view as
USER = None
# if as user fetch contacts and add them to the view
VIEW_CONTACTS = True
# if query with author keys if we are looking for events sent from, sent to, or from and to those keys
DIRECTION = 'both'
# additional profiles to view other than user and anyone they follow
VIEW_EXTRA = None
# look in these 'inboxes' also
INBOXES = None
# max number of events to return init q
LIMIT = 20
# number of hours to look back at star up
SINCE = None
# get events until - note if using until there's not really any point staying running as new events won't come in!
UNTIL = None
# so we can just use a file
CONFIG_FILE = f'view.toml'
# default output event kinds
KINDS = f'{Event.KIND_TEXT_NOTE},{Event.KIND_ENCRYPT}'
# kinds that will be treted as encrypted
ENCRYPT_KINDS = f'{Event.KIND_ENCRYPT}'
# kind that inbox events are expected to be on - currently all have to be on the same kind
# so you'd need to run more than 1 viewer if you want to look into more than one inbox
# and they're using different inbox kinds
INBOX_KINDS = Event.KIND_ENCRYPT
# show only events contained in inboxes (wrapped events)
INBOX_ONLY = False
# for non contacts event must reach this pow to be output
POW = None
# wait for all relays before starting to output events?
START_MODE = 'all'
# how events should be output
OUTPUT = 'formatted'
# when to exit
EXIT = 'never'


def get_int_or_none(val: str, f_name: str) -> int:
    ret = None
    try:
        if val is not None:
            ret = int(val)
    except ValueError as e:
        if val.lower() != 'none':
            raise ConfigError(f'{f_name} - {val} not a numeric value')
    return ret


async def get_from_config(config,
                          profile_handler: ProfileEventHandlerInterface):
    user: Profile = None
    as_sign: SignerInterface = None
    all_view = []
    view_extra = []
    view_contact = []
    inbox_keys = []
    tags = config['tags']
    contacts = config['contacts']

    # keystore for user key aliases
    key_store = get_sqlite_key_store(db_file=config["keystore"]["filename"],
                                     password=config["keystore"]["password"])

    # user we're viewing as
    if config['user'] is not None:
        user_key = (await get_keys_from_str(config['user'],
                                            private_only=False,
                                            key_store=key_store))[0]
        # if we were given a private key then we can create a basic signer
        # so in future this could be something else e.g something external like nsec bunker
        # or hardware device (though that'd probably be pain just for viewing unless it
        # let you batch accept decrypts)
        if user_key.private_key_hex():
            as_sign = BasicKeySigner(user_key)

        user = await profile_handler.aget_profile(user_key.public_key_hex(),
                                                  create_missing=False)
        if user:
            all_view.append(user)
            # lookup and add contacts to view?
            if contacts:
                c_c: Contact
                await profile_handler.aload_contacts(user)
                contacts = user.contacts
                if contacts:
                    contact_ps = await profile_handler.aget_profiles(pub_ks=[c_c.contact_public_key for c_c in contacts],
                                                                     create_missing=True)
                    view_contact = contact_ps.profiles
                    all_view = all_view + contact_ps.profiles
        else:
            print(f'WARNING: unable to find user profile - {config["user"]},can\'t get follows')
            # continue with stubb as_user profile
            user = await profile_handler.aget_profile(user_key.public_key_hex(),
                                                      create_missing=True)
            user.contacts = ContactList(contacts=[],
                                        owner_pub_k=user.public_key)
            all_view.append(user)

    # addtional profiles to view other than current profile
    if config['view_extra']:
        view_keys = (await get_keys_from_str(config['view_extra'],
                                             private_only=False,
                                             key_store=key_store))

        view_ps = await profile_handler.aget_profiles(pub_ks=[k.public_key_hex() for k in view_keys],
                                                      create_missing=True)

        all_view = all_view + view_ps.profiles
        view_extra = view_ps.profiles

    # public inboxes for encrypted messages
    if config['via']:
        # NOTE without user we can only see plain texts in this account
        inbox_keys = (await get_keys_from_str(config['via'],
                                              private_only=True,
                                              key_store=key_store))
        # look up inbox profiles, only done to see if they have a name other than using the pubk
        inboxes = await profile_handler.aget_profiles(pub_ks=[k.public_key_hex() for k in inbox_keys],
                                                      create_missing=True)

        inboxes = inboxes.profiles

        all_view = all_view + inboxes

    # check kinds and encrypt_kinds
    try:
        for k in {'kinds', 'encrypt_kinds'}:
            v = config[k].lower()
            # * for kinds means any
            if k == 'kinds' and v == '*':
                kind_set = None
            # none is valid for encrypt kinds
            elif k == 'encrypt_kinds' and v == 'none':
                kind_set = {}
            else:
                kind_set = {int(v) for v in v.split(',')}

            # update in the config
            config[k] = kind_set
    except ValueError as e:
        raise ConfigError(f'kinds should be integer values got {v}')

    if tags:
        if tags == 'none':
            tags = None
        else:
            tags = tags.split(',')

    # mentioned events
    if config['eid']:
        config['eid'] = config['eid'].split(',')
        for c_id in config['eid']:
            if not Event.is_event_id(c_id):
                raise ConfigError(f'id mentioned event does\'t look like a valid event id {config["kinds"]}')

    config.update({
        'user': user,
        'as_sign': as_sign,
        'all_view': all_view,
        'view_contact': view_contact,
        'view_extra': view_extra,
        # 'inboxes': inboxes,
        'inbox_keys': inbox_keys,
        'since': get_int_or_none(config['since'], 'since'),
        'until': get_int_or_none(config['until'], 'until'),
        'limit': get_int_or_none(config['limit'], 'limit'),
        'tags': tags
    })
    return config


class MyClassifier:

    def __init__(self,
                 event_ids: [str],
                 as_user: Profile,
                 view_contacts: bool,
                 view_profiles: [Profile],
                 public_inboxes: [Profile]):

        self._as_user = as_user
        self._view_contacts = view_contacts
        self._view_profiles = view_profiles
        self._inboxes = public_inboxes
        self._view_keys = []
        self._event_ids = event_ids

    async def make_ready(self):
        await self._make_view_keys()

    @property
    def view_keys(self):
        return self._view_keys

    async def _make_view_keys(self):
        c_c: Contact
        c_p: Profile
        c_i: Inbox

        if self._as_user is not None:
            self._view_keys.append(self._as_user.public_key)
            if self._view_contacts:
                self._view_keys = self._view_keys + [c_c.contact_public_key for c_c in self._as_user.contacts]

        if self._view_profiles is not None:
            self._view_keys = self._view_keys + [c_p.public_key for c_p in self._view_profiles]

        if self._inboxes is not None:
            self._view_keys = self._view_keys + [await c_i.pub_key for c_i in self._inboxes]

    def classify(self, evt: Event):
        # we requested it but it might still be considered spam, we mark here
        # if maybe spam is returned then extra test on event might happen
        # currently nip5 check
        ret = 'maybe spam'

        # we specifically asked for this event
        if self._event_ids and evt.id in self._event_ids:
            ret = 'good'

        # event pub_k it to or from keys we have specificly mentioned
        elif evt.pub_key in self._view_keys or \
                self._as_user is not None and \
                (self._as_user.public_key in evt.pub_key or self._as_user.public_key in evt.p_tags):
            ret = 'good'

        return ret


class PrintEventHandler(EventHandler):
    """
       print handler
       printer should be a class with a print_event(Event) method
    """
    def __init__(self,
                 event_acceptors=[],
                 profile_handler: ProfileEventHandlerInterface = None,
                 printer=None,
                 classifier=None,
                 nip5helper: NIP5Helper = None,
                 nip5=False,
                 pow=False):

        self._profile_handler = profile_handler
        self._printer = printer
        self._classifier: MyClassifier = classifier
        self._nip_checker: NIP5Helper = nip5helper
        self._pow = pow
        self._nip5 = nip5

        # prints are queued
        self._print_queue = asyncio.Queue()
        asyncio.create_task(self._my_consumer())

        super().__init__(event_acceptors)

    async def _valid_nip5_event_pub(self, evt: Event) -> bool:
        ret = False
        p: Profile = await self._profile_handler.aget_profile(evt.pub_key)
        try:
            ret = await self._nip_checker.is_valid_profile(p)
        except NIP5Error as ne:
            logging.debug(f'PrintEventHandler::_printer_event_if_nip5 ignored event {evt.id} publisher has bad nip5')
        return ret

    async def _profiles_prefetch(self, events: [Event]):
        ukeys = set()
        for c_evt in events:
            ukeys.add(c_evt.pub_key)
            for c_tag in c_evt.p_tags:
                ukeys.add(c_tag)
        await self._profile_handler.aget_profiles(list(ukeys),
                                                  create_missing=True)

    async def ado_event(self, the_client: Client, sub_id, evt: list[Event]|Event):
        c_evt: Event
        if isinstance(evt, Event):
            evt = [evt]

        to_print = [
            c_evt for c_evt in evt if self.accept_event(the_client, sub_id, c_evt)
        ]
        if not to_print:
            return

        Event.sort(to_print, inplace=True, reverse=False)
        await self._profiles_prefetch(to_print)

        for c_evt in to_print:
            if self._classifier:
                rating = self._classifier.classify(c_evt)
                # good events will output always
                # maybe spam events will pass
                #   if pow and nip5 are both false
                #   if pow is true and nip5 is false
                #   if pow is false and nip5 is true and nip5 validates
                #   if pow is true and nip5 is true and nip5 validates

                do_print = rating == 'good' or \
                           (self._pow is False and self._nip5 is False) or \
                           (self._pow and self._nip5 is False) or \
                           (self._nip5 and await self._valid_nip5_event_pub(c_evt))

                if do_print:
                    await self._printer.aprint_event(the_client, sub_id, c_evt)

    async def astatus(self, status):
        await self._printer.astatus(status)

    def do_event(self, the_client: Client, sub_id, evt: Event):
        # add the the evt to the print queue
        self._print_queue.put_nowait((the_client, sub_id, evt))

    async def _my_consumer(self):
        try:
            while True:
                args = await self._print_queue.get()
                await self.ado_event(*args)
        except Exception as e:
            logging.debug(f'PrintEventHandler::_my_consumer - {e}')


def get_args() -> dict:
    """
    get args to use order is
        default -> toml_file -> cmd_line options

    so command line option is given priority if given

    :return: {}
    """

    ret = {
        'work_dir': WORK_DIR,
        'conf': CONFIG_FILE,
        'relay': RELAYS,
        'user': USER,
        'as_sign': None,
        'contacts': VIEW_CONTACTS,
        'view_extra': VIEW_EXTRA,
        'via': INBOXES,
        'direction': DIRECTION,
        'limit': LIMIT,
        'since': SINCE,
        'until': UNTIL,
        'kinds': KINDS,
        'encrypt_kinds': ENCRYPT_KINDS,
        'inbox_kinds': INBOX_KINDS,
        'inbox_only': INBOX_ONLY,
        'pow': POW,
        'nip5': False,
        'nip5check': False,
        'pubkey': False,
        'match_tags': None,
        'hashtag': None,
        'tags': None,
        'eid': None,
        'start_mode': START_MODE,
        'output': OUTPUT,
        'ssl_disable_verify': None,
        'entities': False,
        'exit': EXIT,
        'keystore': {
            'filename': WORK_DIR + KEY_STORE_DB_FILE,
            'password': None
        },
        'debug': False,
    }

    # only done to get the work-dir and conf options if set
    ret.update(get_cmdline_args(ret))
    # now form config file if any
    # now form config file if any
    load_toml(filename=ret['conf'],
              dir=ret['work_dir'],
              current_args=ret)

    # 2pass so that cmdline options override conf file options
    ret.update(get_cmdline_args(ret))

    # if debug flagged enable now
    if ret['debug'] is True:
        logging.getLogger().setLevel(logging.DEBUG)
        logging.debug(f'get_args:: running with options - {ret}')

    if not ret['relay']:
        ret['relay'] = input('relays: ')

    if ret['inbox_only'] and ret['via'] is None:
        ret['via'] = input('inboxes: ')

    return ret


def get_cmdline_args(args) -> dict:
    parser = argparse.ArgumentParser(
        prog='view.py',
        description="""
            view nostr events from the command line
            """
    )
    parser.add_argument('-c', '--conf', action='store', default=args['conf'],
                        help=f'name com TOML file to use for configuration, default[{args["conf"]}]')
    parser.add_argument('--work-dir', action='store', default=args['work_dir'],
                        help=f'base dir for files used if full path isn\'t given, default[{args["work_dir"]}]')
    parser.add_argument('-r', '--relay', action='store', default=args['relay'],
                        help=f'comma separated nostr relays to connect to, default[{args["relay"]}]')
    parser.add_argument('-u', '--user', action='store', default=args['user'],
                        help=f"""
                        alias, priv_k or pub_k of user to view as. If only created from pub_k then kind 4
                        encrypted events will be left encrypted, 
                        default[{args['user']}]""")
    parser.add_argument('--contacts', action='store_true',
                        help='if --as-user lookup contacts and add to view',
                        default=args['contacts'])
    parser.add_argument('--no-contacts', action='store_false',
                        help='if --as-user DO NOT add contacts to view',
                        default=args['contacts'], dest='contacts')
    parser.add_argument('--view-extra', action='store', default=args['view_extra'],
                        help=f"""
                            additional comma separated alias, priv_k or pub_k of user to view,
                            default[{args['view_extra']}]""")
    parser.add_argument('-v', '--via', action='store', default=args['via'],
                        help=f"""
                            additional comma separated alias(with priv_k) or priv_k that will be used 
                            as public inbox with wrapped events,
                            default[{args['via']}]""")
    parser.add_argument('--direction', action='store', default=args['direction'],choices={'from', 'to', 'both'},
                        help=f"""
                                if query with author keys if we are looking for events sent from, sent to or both with those keys
                                default[{args['direction']}]""")
    parser.add_argument('-i', '--id', action='store', default=args['eid'], dest='eid',
                        help=f"""
                                    comma separated event ids will be added as e tag filter e.g with kind=42 
                                    can be used to view a chat channel,
                                    default[{args['eid']}]""")
    parser.add_argument('-k', '--kinds', action='store', default=args['kinds'],
                        help=f"""
                                comma separated event kinds to output,
                                default[{args['kinds']}]""")
    parser.add_argument('--encrypt-kinds', action='store', default=args['encrypt_kinds'],
                        help=f"""
                                    comma separated event kinds to be decrypted,
                                    default[{args['encrypt_kinds']}]""")
    parser.add_argument('--inbox-kinds', action='store', default=args['inbox_kinds'], type=int,
                        help=f"""
                                    kind to use for inbox events, applied to all inboxes
                                    default[{args['inbox_kinds']}]""")
    parser.add_argument('-l', '--limit', action='store', default=args['limit'],
                        help=f'max number of events to return, default [{args["limit"]}]')
    parser.add_argument('-s', '--since', action='store', default=args['since'],
                        help=f'show events n hours previous to running, default [{args["since"]}]')
    parser.add_argument('--until', action='store', default=args['until'],
                        help=f'show events n hours after since, default [{args["until"]}]')
    parser.add_argument('--hashtag', action='store', default=args['hashtag'],
                        help=f'only events with t tag value will be matched, default[{args["hashtag"]}]')
    parser.add_argument('--pubkey', action='store_true', default=args['pubkey'],
                        help=f"""
                                        output event author pubkey
                                        default[{args['pubkey']}]""")
    parser.add_argument('-t', '--tags', action='store', default=args['tags'],
                        help=f"""
                                    comma separated tag types to output, =* for all
                                    default[{args['tags']}]""")

    parser.add_argument('-p', '--pow', action='store', choices=[8, 12, 16, 20, 24, 28, 32], default=args['pow'],
                        type=int,
                        help=f"""
                                        minimum amount required for events excluding contacts of user
                                        default[{args['pow']}]""")

    parser.add_argument('-e', '--entities', action='store_true',
                        help='output event_id and pubkeys as nostr entities',
                        default=args['entities'])
    parser.add_argument('--no-entities', action='store_false',
                        help='do not output event_id and pubkeys as nostr entities',
                        default=(not args['entities']), dest='entities')
    parser.add_argument('--nip5check', action='store_true',
                        help='nip5 checked and displayed green if valid',
                        default=args['nip5check'])
    parser.add_argument('-n', '--nip5', action='store_true', help='valid nip5 required for events excluding contacts of user',
                        default=args['nip5'])
    parser.add_argument('--start-mode', choices=['all', 'first'],
                        default=args['start_mode'],
                        help=f"""
                                    at start wait for ALL relays to return events before starting to print or just FIRST 
                                    default[{args['start_mode']}]""")

    parser.add_argument('--inbox-only', action='store_true', help='only show events that are contained in inboxes',
                        default=args['inbox_only'])
    parser.add_argument('--no-inbox-only', action='store_false', help='events in and outside of inboxes will be shown',
                        default=args['inbox_only'])
    parser.add_argument('-o', '--output', action='store', choices=['formatted', 'json', 'content'], default=args['output'],
                        help=f"""
                                        how to display events
                                        default[{args['output']}]""")
    parser.add_argument('--ssl-disable-verify', action='store_true', help='disables checks of ssl certificates')
    # TODO add until option - if have until exit when we get event with timestamp >= that
    parser.add_argument('-x', '--exit', action='store',
                        default=args['exit'], choices=['never', 'store'],
                        help=f"""
                                            never - run indefinitely. store - exit after receiving stored events. 
                                            default[{args['exit']}]""")
    parser.add_argument('-d', '--debug', action='store_true', help='enable debug output', default=args['debug'])

    ret = parser.parse_args()
    # so --as_user opt can be overridden empty if its defined in config file
    if ret.user and (ret.user == '' or ret.user.lower() == 'none'):
        ret.user = None

    return vars(ret)


async def get_event_filters(view_profiles: [Profile],
                      since: datetime,
                      until: datetime,
                      limit: int,
                      mention_eids: [str],
                      kinds: [int],
                      hash_tag: str,
                      pow: int,
                      nip5: bool,
                      direction: str,
                      inboxes: [Inbox],
                      inbox_only: bool):

    ret = []
    watch_keys = []
    c_p: Profile
    c_i: Inbox

    # add the non-inbox view event filter
    if inbox_only is False:
        if view_profiles:
            for c_p in view_profiles:
                watch_keys.append(c_p.public_key)

        # watch both from and mention for these keys
        if watch_keys:
            # events from accounts we follow, pow if any, not applied
            if direction in {'both', 'from'}:
                ret.append({
                    'authors': watch_keys
                })

            if direction in {'both', 'to'}:

                ret.append({
                    '#p': watch_keys
                })

        # either not watching any particular authors or
        # we're going to allow unrequested events through based on
        # pow, nip5 or pow and nip5
        # note in case of nip5 alone you have to fetch all events anyhow so
        # probably best to use with at least a minimal pow
        if not watch_keys or pow or nip5:
            ret.append({})

    # add view into each inbox
    if inboxes:
        ret.append({
            # all have to be on same kind
            'kinds': [inboxes[0].kind],
            # inboxes are actually profile objs not inboxes here :(!
            'authors': [await c_i.pub_key for c_i in inboxes]
        })

    # add common filter paras to all filters
    for c_f in ret:
        # kinds addded unless we already added because its an inbox
        if kinds is not None and 'kinds' not in c_f:
            c_f['kinds'] = list(kinds)

        # other fields that may be added
        if limit:
            c_f['limit'] = limit
        if since:
            c_f['since'] = util_funcs.date_as_ticks(since)
        if until:
            c_f['until'] = util_funcs.date_as_ticks(until)
        if mention_eids:
            c_f['#e'] = mention_eids
        if hash_tag:
            c_f['#t'] = [hash_tag]

        # pow added if we didn't specify authors, this include to filter
        # (pow is applied to events sent/mentioning authors we asked for)
        if pow and 'authors' not in c_f:
            c_f['ids'] = [''.join(['0'] * (int(pow / 4)))]

    # any requested eventids we'll fetch regardless of by who or when
    if mention_eids:
        ret.append({
            'ids': mention_eids
        })

    return ret


def output_rel_date(rel_date: datetime, the_date: datetime) -> str:
    """
        returns str output for the_date base on rel_date
        at the moment this just means if rel_date.date() == the_date.date()
        date output won't be output
    """
    if rel_date.date() == the_date.date():
        ret = the_date.strftime('%H:%M:%S')
    else:
        ret = the_date.strftime('%d/%m/%Y-%H:%M:%S')
    return ret


async def print_run_info(relay,
                         as_user,
                         contacts,
                         extra_view_profiles,
                         inboxes,
                         view_kinds,
                         encrypted_kinds,
                         since,
                         until,
                         limit,
                         pow,
                         nip5,
                         profile_handler,
                         direction,
                         inbox_only,
                         ):
    c_p: Profile
    c_c: Contact
    c_i: Inbox
    print(f'using relays {relay}')

    # output running info
    if as_user:
        print('events will be displayed as user %s' % as_user.display_name())
        if contacts:
            print('--- contacts ---')

            # this will group fetch all follow profiles so they won't be fetch individually
            # when we list
            await profile_handler.aget_profiles(pub_ks=as_user.contacts.follow_keys(),
                                                create_missing=True)

            for f_k in as_user.contacts.follow_keys()[:10]:
                c_p = await profile_handler.aget_profile(f_k)
                if c_p:
                    print(c_p.display_name())
                else:
                    print(c_c.contact_public_key)
            if len(as_user.contacts) > 10:
                print(f'and {len(as_user.contacts)-10} more...')
        else:
            print('contacts not added to view')

    else:
        print('runnning without a user')

    if extra_view_profiles:
        print('--- extra profiles ---')
        for c_p in extra_view_profiles:
            print(c_p.display_name())

    if inboxes:
        print(f'--- checking inboxes using kind {inboxes[0].kind} ---')
        for c_i in inboxes:
            print(await c_i.name)

    if view_kinds:
        filter_about = [f'showing events of kind {view_kinds}']
    else:
        filter_about = [f'showing events of kind any']

    if encrypted_kinds:
        filter_about.append(f' encrypted kinds {encrypted_kinds}')

    now = datetime.now()
    if since is not None:
        filter_about.append(f' from {output_rel_date(now, since)}')

    if until:
        filter_about.append(f' -> {output_rel_date(since, until)}')

    filter_about.append(f' limit {limit}')

    print(''.join(filter_about))

    print(f'pow for non follows ({pow}) and nip5 check ({nip5}) direction ({direction})')
    if inbox_only:
        print('!! inbox only !!')


async def main(args):
    # connect to these relay
    relay = args['relay'].split(',')

    # disable ssl checks
    ssl = None
    if args['ssl_disable_verify']:
        ssl = False

    # start and connect relays
    my_client = ClientPool(relay, ssl=ssl)
    asyncio.create_task(my_client.run())
    # await my_client.wait_connect(timeout=10,)

    profile_handler = NetworkedProfileEventHandler(client=my_client)

    # fills in config with data from nostr, e.g. follows for any user
    try:
        config = await get_from_config(args, profile_handler)
    except ConfigError as ce:
        print(ce)
        my_client.end()
        sys.exit(2)
    except Exception as e:
        print(e)
        my_client.end()
        sys.exit(2)

    # user that we running as, if this user has priv_k then where needed we'll do decryption and output plaintext
    user: Profile = config['user']
    as_sign = config['as_sign']

    # add add user contacts to view
    c_p: Profile
    view_contacts = config['contacts']
    view_contacts_profiles = config['view_contact']
    view_contacts_k = [c_p.keys for c_p in view_contacts_profiles]

    # extra profiles requested to view other than contacts of user or inboxes
    extra_view_profiles = config['view_extra']
    extra_view_k = [c_p.keys for c_p in extra_view_profiles]

    # just the keys of above, in most cases the keys are all we want - this is what we pass into outputters
    # for when they create the maps with inbox wrap events
    # probably we could put all ks in and use the same as we use on teh main filter but probably its not necessary
    # e.g. would you expect to use inbox and then look in that inbox for encrypted msgs with that inbox k?
    # everyone who has teh inbox k could look into them anyway so why?
    extra_and_contact_k = view_contacts_k + extra_view_k

    # all view profiles, contains all pub_ks that we're going to request from relay to make our view
    view_profiles = config['all_view']

    # the event kinds that we subscribe to and output
    view_kinds = config['kinds']

    # the event kinds that we will decrypt (we only add those that appear in view kinds)
    if view_kinds:
        encrypt_kinds = config['encrypt_kinds'].intersection(view_kinds)
    else:
        encrypt_kinds = config['encrypt_kinds']

    # the kind we'll use when looking for inbox events - same for all inboxes
    inbox_kinds = config['inbox_kinds']
    # in inbox only only events wrapped in given inboxes will be output
    inbox_only = config['inbox_only']

    # actually make into inbox objects
    inboxes = []

    inbox_keys: [Keys] = config['inbox_keys']
    c_k: Keys
    for c_k in inbox_keys:
        n_inbox = Inbox(signer=BasicKeySigner(c_k),
                        name=profile_handler.get_profile(c_k.public_key_hex()).display_name(),
                        use_kind=inbox_kinds)

        inboxes.append(n_inbox)

    # now set share map on inboxes if so we can see/decrypt encrypted message in them
    if as_sign:
        for c_i in inboxes:
            await c_i.set_share_map(for_sign=as_sign,
                                    to_keys=extra_and_contact_k)

    # sent from or to or both keys we've given
    direction = config['direction']

    # on startup wait for all relys before starting printout?
    start_mode = config['start_mode']

    # if this is true then output is just the json as we recieve ot
    output = config['output']

    # bits of pow to events from any accounts we didn't request
    pow = config['pow']

    # nip5 required for events from any accounts we didn't request
    # if both pow and nip5 are requested then both needed for those events to be output
    nip5 = config['nip5']
    nip5check = config['nip5check']

    # used if we are spam checking with nip5 or if we are verifying nip5s for display
    nip5helper = NIP5Helper()

    # only view events that mention this event
    mention_eids = config['eid']

    # actually show event author pubkey
    show_pubkey = config['pubkey']

    # output these event tags
    show_tags = config['tags']

    # output the event id and pubkey as note or npub style
    entities = config['entities']

    # keep track of last events seen so we can reconnect without getting same events again
    last_event_track = LastEventHandler()

    # events from since point, after initial events this is updated
    # so on reconnect we don't go back and get all old events again
    since = config['since']
    if since is not None:
        since = datetime.now() - timedelta(hours=since)

    # until this date - note on reaching until date (maybe in the future)
    # we might as well exit or have option to exit as no more events will come in
    until = config['until']
    if until is not None:
        until = since + timedelta(hours=until)

    # if given only events with this t tag will be returned
    hash_tag = config['hashtag']

    # rough max limit of events at bootstrap (before live incoming events)
    limit = config['limit']

    # used to check that the events that we get back from relay match those we requested
    # the actual filter used is set when we add the filter in on_connect
    e_filter_acceptor = FilterAcceptor()

    # exit mode
    exit = config['exit']

    # just a wrap around get_event_filters so we don't have to call with all the
    # fields all the time
    async def my_get_event_filters(with_since):

        return await get_event_filters(view_profiles=view_profiles,
                                       since=with_since,
                                       until=until,
                                       limit=limit,
                                       mention_eids=mention_eids,
                                       kinds=view_kinds,
                                       pow=pow,
                                       nip5=nip5,
                                       direction=direction,
                                       hash_tag=hash_tag,
                                       inboxes=inboxes,
                                       inbox_only=inbox_only)

    # show run info except if we're outputting json
    if output != 'json':
        await print_run_info(relay=relay,
                             as_user=user,
                             contacts=view_contacts,
                             extra_view_profiles=extra_view_profiles,
                             inboxes=inboxes,
                             direction=direction,
                             view_kinds=view_kinds,
                             encrypted_kinds=encrypt_kinds,
                             since=since,
                             until=until,
                             limit=limit,
                             pow=pow,
                             nip5=nip5,
                             profile_handler=profile_handler,
                             inbox_only=inbox_only)

    def my_connect(the_client: Client):
        # so on reconnect we don't ask for everything again
        use_since = since

        if last_event_track.get_last_event_dt(the_client):
            use_since = last_event_track.get_last_event_dt(the_client)+timedelta(seconds=1)

        # sub just for keeping profiles upto date
        the_client.subscribe(handlers=[profile_handler,
                                       last_event_track],
                             filters=[
            {
                'kinds': [Event.KIND_META],
                'since': util_funcs.date_as_ticks(datetime.now())
            }
        ])

        asyncio.create_task(do_event_sub(the_client, use_since))

    async def do_event_sub(the_client: Client, use_since):
        # get the event filter with since
        event_filter = await my_get_event_filters(with_since=use_since)
        # set filter to the acceptor - if a relay gives us events that don't pass our filter we'll reject
        e_filter_acceptor.filter = event_filter
        # now subscribe
        the_client.subscribe(handlers=[last_event_track, print_handler],
                             filters=event_filter)

    # actually does the outputting
    if output == 'json':
        my_printer = JSONPrinter(as_sign=as_sign,
                                 inboxes=inboxes,
                                 encrypted_kinds=encrypt_kinds)
    elif output == 'formatted':
        # by default we won't valid the nip5s for display
        fnip_check = None
        if nip5check:
            fnip_check = nip5helper

        my_printer = FormattedEventPrinter(profile_handler=profile_handler,
                                           as_sign=as_sign,
                                           as_user=user,
                                           inboxes=inboxes,
                                           show_pub_key=show_pubkey,
                                           show_tags=show_tags,
                                           entities=entities,
                                           nip5helper=fnip_check,
                                           kinds=view_kinds,
                                           encrypted_kinds=encrypt_kinds)
    elif output == 'content':
        my_printer = ContentPrinter(as_sign=as_sign,
                                    inboxes=inboxes,
                                    encrypted_kinds=encrypt_kinds)

    # initial filter to get upto now
    boot_e_filter = await my_get_event_filters(with_since=since)

    # set inital accept criteria, this is a double check on what relays return
    # e_filter_acceptor.filter = boot_e_filter

    #
    my_classifier = MyClassifier(event_ids=mention_eids,
                                 as_user=user,
                                 view_contacts=view_contacts,
                                 view_profiles=view_profiles,
                                 public_inboxes=inboxes)
    await my_classifier.make_ready()
    # event handler for printing events
    print_handler = PrintEventHandler(profile_handler=profile_handler,
                                      event_acceptors=[DeduplicateAcceptor(),
                                                       e_filter_acceptor,
                                                       LengthAcceptor(),
                                                       NotOnlyNumbersAcceptor()
                                                       ],
                                      printer=my_printer,
                                      classifier=my_classifier,
                                      nip5helper=nip5helper,
                                      pow=pow is not None,
                                      nip5=nip5)

    """
        if order_first, then we wait to get events from each relay and then order them as a single set
        otherwise a slower relay returning later might make the events lookout of order date wise
        (note that its always possible to get out of date events but normally they should come in
        more or less forward in dates)
        without order first events will start printing as soon as the fastest relay returns
    """
    keep_running = True
    if start_mode == 'all':
        the_events = await my_client.query(filters=boot_e_filter,
                                           do_event=last_event_track.do_event,
                                           emulate_single=True,
                                           wait_connect=False)

        # because events may come from mutiple relays  sort
        Event.sort(the_events, inplace=True)

        # merge from multiple relays means that we can get more events than the limit
        # (different events from different relays - add opt to disable this force cut to limit?)
        if limit is not None:
            the_events = the_events[:limit]

        await print_handler.ado_event(
            the_client=None,
            sub_id=None,
            evt=the_events)

        keep_running = exit == 'never'
        if keep_running:
            await my_printer.astatus('*** listening for more events ***')

    else:

        def adhoc_do_event(the_client: Client, sub_id: str, events: [Event]):
            # events drawn in and they're recieved from each relay (minus duplicates)
            asyncio.create_task(print_handler.ado_event(
                the_client=None,
                sub_id=None,
                evt=events))
            last_event_track.do_event(the_client, sub_id, events)

        # called at query completetion - which might not be until timeout if we have bad relays
        def first_pull_complete(events: [Event]):
            nonlocal keep_running
            keep_running = exit == 'never'
            if keep_running:
                asyncio.create_task(my_printer.astatus('*** listening for more events ***'))

        # TODO: fix client so that do event can be [handlers] and we wouldnt need adhoc_do_event
        await my_client.query(filters=boot_e_filter,
                              do_event=adhoc_do_event,
                              on_complete=first_pull_complete,
                              wait_connect=True,
                              emulate_single=False)

    for c_client in my_client:
        my_client.set_on_eose(print_handler.do_event)
        # because we're already connected we'll call manually
        if c_client.connected:
            my_connect(c_client)
            my_client.set_on_connect(my_connect)
    while keep_running:
        await asyncio.sleep(0.1)
    my_client.end()


if __name__ == "__main__":
    logging.getLogger().setLevel(logging.ERROR)
    util_funcs.create_work_dir(WORK_DIR)
    def sigint_handler(signal, frame):
        sys.exit(0)

    signal.signal(signal.SIGINT, sigint_handler)
    asyncio.run(main(get_args()))



import logging
import sys
import os
import asyncio
import signal
import argparse
from copy import copy
from pathlib import Path
import shutil
try:
    from stem.control import Controller
except Exception as e:
    pass
from monstr.relay.relay import Relay
from monstr.relay.accept_handlers import LengthAcceptReqHandler, CreateAtAcceptor, AuthenticatedAcceptor
from monstr.relay.tor import TORService
from monstr.event.persist_postgres import RelayPostgresEventStore
from monstr.event.persist_sqlite import ARelaySQLiteEventStore
from monstr.event.persist_memory import RelayMemoryEventStore
from monstr.event.expire import ASQLiteNIP40Expirer, MemoryNIP40Expirer
from monstr.util import ConfigError
from monstr_terminal.util import load_toml
import ssl

# default values when nothing is specified either from cmd line or config file
HOST = 'localhost'
PORT = 8081
END_POINT = '/'
DEBUG_LEVEL = logging.DEBUG
DB_TYPE = 'sqlite'
# make this default home, wouldn't work on windows
WORK_DIR = f'{Path.home()}/.nostrpy/'
CONFIG_FILE = 'relay.toml'
SQL_LITE_FILE = f'{WORK_DIR}nostr-relay.db'
PG_USER = 'postgres'
PG_PASSWORD = 'password'
PG_DATABASE = 'nostr-relay'
MAX_SUB = 10
MAX_CONTENT_LENGTH = None
SSL = False
# acceptance created_at (mins)
MAX_BEFORE = None
MAX_AFTER = 5

def create_work_dir():
    if not os.path.isdir(WORK_DIR):
        logging.info(f'create_work_dir:: attempting to create {WORK_DIR}')
        os.makedirs(WORK_DIR)


async def get_sql_store(filename, is_nip16, is_nip33):
    f = Path(filename)

    parent_dir = f.parts[len(f.parts)-2]

    # user must have given another dir, we better check it exists...
    if parent_dir != '.nostrpy':
        my_dir = Path(os.path.sep.join(f.parts[:-1]).replace(os.path.sep+os.path.sep, os.path.sep))
        if not my_dir.is_dir():
            print(f'sqllite dir not found {my_dir}')
            sys.exit(2)

    # if the file doesn't exist it'll be created and we'll create the db struct too
    # if it does we'll assume everything is ok...we could do more

    ret = ARelaySQLiteEventStore(filename,
                                 is_nip16=is_nip16,
                                 is_nip33=is_nip33)
    if not ret.exists():
        logging.info(f'get_sql_store::create new db {filename}')
        await ret.create()
    else:
        logging.info(f'get_sql_store::open existing db {filename}')

    return ret


def get_postgres_store(db_name, user, password, is_nip16, is_nip33):
    ret = RelayPostgresEventStore(db_name=db_name,
                                  user=user,
                                  password=password,
                                  is_nip16=is_nip16,
                                  is_nip33=is_nip33)

    if not ret.exists():
        ret.create()
    return ret


def get_cmdline_args(args) -> dict:
    parser = argparse.ArgumentParser(
        prog='run_relay.py',
        description="""
            runs a nostr relay
            """
    )
    parser.add_argument('--host', action='store', default=args['host'],
                        help=f'ip address where relay will listen, default[{args["host"]}]')
    parser.add_argument('--port', action='store', default=args['port'], type=int,
                        help=f'port relay will listen, default[{args["port"]}]')
    parser.add_argument('--endpoint', action='store', default=args['endpoint'],
                        help=f'endpoint address for the relay websocket[{args["endpoint"]}]')
    parser.add_argument('-s', '--store', action='store', default=args['store'],
                        choices=['sqlite', 'postgres', 'transient', 'none'],
                        help=f'storage type to use for received events, default[{args["store"]}]')

    # sqlite store stuff
    parser.add_argument('--dbfile', action='store', default=args['dbfile'],
                        help=f'when store is sqlite the file location for the db, default[{args["dbfile"]}]')
    # postgres store stuff
    parser.add_argument('--pg_database', action='store', default=args['pg_database'],
                        help=f'when store is postgres the postgres db name, default[{args["pg_database"]}]')
    parser.add_argument('--pg_user', action='store', default=args['pg_user'],
                        help=f'when store is postgres the postgres username, default[{args["pg_user"]}]')
    parser.add_argument('--pg_password', action='store', default=args['pg_password'],
                        help=f'when store is postgres the postgres password')
    # general relay operation

    parser.add_argument('--maxsub', action='store', default=args['maxsub'], type=int,
                        help=f'maximum open subs allowed per client websocket, default[{args["maxsub"]}]')
    parser.add_argument('--maxlength', action='store', default=args['maxlength'], type=int,
                        help=f'maximum length for event content if any, default[{args["maxlength"]}]')

    # nip22 acceptable created_at ranges
    parser.add_argument('--max-before', action='store', default=args['max_before'], type=int,
                        help=f'maximum time before current time to accept created_at of events if any (mins), default[{args["max_before"]}]')
    parser.add_argument('--max-after', action='store', default=args['max_after'], type=int,
                        help=f'maximum time after current time to accept created_at of events if any (mins), default[{args["max_after"]}]')

    # nip support flags, both enable and disable versions so a def true can be overriden
    parser.add_argument('--nip16', action='store_true',
                        help=f"""enable NIP16 - Event treatment, ephemeral and replaceable event ranges
                see https://github.com/nostr-protocol/nips/blob/master/16.md, default[{args["nip16"]}]""",
                        default=args['nip16'])

    parser.add_argument('--no-nip16', action='store_false',dest='nip16',
                        help=f"""disable NIP16, default[{not args["nip16"]}]""",
                        default=args['nip16'])

    parser.add_argument('--nip20', action='store_true',
                        help=f"""enable NIP20 - OK command events  
                see https://github.com/nostr-protocol/nips/blob/master/20.md, default[{args["nip20"]}]""",
                        default=args['nip20'])
    parser.add_argument('--no-nip20', action='store_false',dest='nip20',
                        help=f"""disable NIP20, default[{not args["nip20"]}]""",
                        default=args['nip20'])

    parser.add_argument('--nip33', action='store_true',
                        help=f"""enable NIP33 - Parameterized Replaceable Events
                    see https://github.com/nostr-protocol/nips/blob/master/20.md, default[{args["nip33"]}]""",
                        default=args['nip33'])
    parser.add_argument('--no-nip33', action='store_false',dest='nip33',
                        help=f"""disable NIP33, default[{not args["nip33"]}]""",
                        default=args['nip33'])

    parser.add_argument('--nip40', action='store_true',
                        help=f"""enable NIP40 - Expiration Timestamp
                        see https://github.com/nostr-protocol/nips/blob/master/40.md, default[{args["nip40"]}]""",
                        default=args['nip40'])
    parser.add_argument('--no-nip40', action='store_false', dest='nip40',
                        help=f"""disable NIP40, default[{not args["nip40"]}]""",
                        default=args['nip40'])

    # TODO: add extra support file list of keys, or to use list event by given pub_k to set access keys
    parser.add_argument('--nip42', action='store_true',
                        help=f"""enable NIP42 - Authentication of clients to relays
                            see https://github.com/nostr-protocol/nips/blob/master/42.md, default[{args["nip42"]}]""",
                        default=args['nip42'])
    parser.add_argument('--no-nip42', action='store_false', dest='nip42',
                        help=f"""disable NIP42, default[{not args["nip42"]}]""",
                        default=args['nip42'])

    # end nips
    parser.add_argument('--ssl', action='store_true', help='run ssl ssl_key and ssl_cert will need to be defined',
                        default=args['ssl'])
    parser.add_argument('--tor', action='store_true', help='make realy accessable over tor',
                        default=args['tor'])
    parser.add_argument('-w', '--wipe', action='store_true', help='wipes event store and exits', default=False)
    parser.add_argument('-d', '--debug', action='store_true', help='enable debug output', default=args['debug'])

    ret = parser.parse_args()

    return vars(ret)


def get_args() -> dict:
    """
    get args to use order is
        default -> toml_file -> cmd_line options

    so command line option is given priority if given

    :return: {}
    """

    # set up the defaults if not overriden
    # add config option so we can have different files and also make workdir conf?
    ret = {
        'host': HOST,
        'port': PORT,
        'endpoint': END_POINT,
        'store': DB_TYPE,
        'dbfile': SQL_LITE_FILE,
        'pg_database': PG_DATABASE,
        'pg_user': PG_USER,
        'pg_password': PG_PASSWORD,
        'maxsub': MAX_SUB,
        'maxlength': MAX_CONTENT_LENGTH,
        'max_before': MAX_BEFORE,
        'max_after': MAX_AFTER,
        'nip16': True,
        'nip20': True,
        'nip33': True,
        'nip40': True,
        'nip42': False,
        'ssl': SSL,
        'ssl_key': None,
        'ssl_cert': None,
        'tor': False,
        'tor_password': None,
        'tor_service_dir': None,
        'tor_empheral': True,
        'debug': False
    }

    # now form config file if any
    load_toml(filename=CONFIG_FILE,
              dir=WORK_DIR,
              current_args=ret)

    # now from cmd line
    ret.update(get_cmdline_args(ret))

    # if debug flagged enable now
    if ret['debug'] is True:
        logging.getLogger().setLevel(logging.DEBUG)

    # remove any config not required because of storage choice
    if ret['store'] != 'sqlite':
        del ret['dbfile']
    if ret['store'] != 'postgres':
        del ret['pg_database']
        del ret['pg_user']
        del ret['pg_password']

    # don't out the password
    ret_out = copy(ret)
    if 'pg_password' in ret:
        ret_out['pg_password'] = '****'

    if ret['debug']:
        logging.debug(f'get_args:: running with options - {ret}')

    if ret['ssl']:
        if ret['ssl_key'] is None or ret['ssl_cert'] is None:
            raise ConfigError('both ssl_key and ssl_cert options must be provide to run as ssl')

    return ret


async def main(args):
    is_wipe = args['wipe']
    create_work_dir()

    # relay addressing
    host = args['host']
    port = args['port']
    end_point = args['endpoint']

    # sub options
    max_sub = args['maxsub']
    max_length = args['maxlength']

    # nip22 created out acceptable
    max_before = args['max_before']
    if max_before:
        max_before = max_before * 60
    max_after = args['max_after']
    if max_after:
        max_after = max_after * 60

    # get nip flags
    nip16 = args['nip16']
    nip20 = args['nip20']
    nip33 = args['nip33']
    nip40 = args['nip40']
    nip42 = args['nip42']


    # ssl options
    is_ssl = args['ssl']
    ssl_cert = args['ssl_cert']
    ssl_key = args['ssl_key']

    # relay information NIP11
    relay_info = {
        'software': 'https://github.com/monty888/monstr_terminal',
        'version': '0.1.1'
    }
    # copy fields from args - from TOML file
    if 'relay_information' in args:
        for k, v in args['relay_information'].items():
            # only these fields will be added
            if k in {'name', 'pubkey', 'contact'}:
                relay_info[k] = v


    # empheral is easier as doesn't need to access any files
    # password, service_dir and empheral can only be se in toml currently
    enable_tor = args['tor']
    tor_password = args['tor_password']
    tor_dir = args['tor_service_dir']
    tor_empheral = args['tor_empheral']

    # get the store type we're using and create
    store = args['store']
    my_store = None
    my_expire = None
    # if expiration enable, events removed every 5 mins
    expire_interval = 60*5

    # create storage object which is either to sqllite, posgres or transient
    if store == 'sqlite':
        my_store = await get_sql_store(filename=args['dbfile'],
                                       is_nip16=nip16,
                                       is_nip33=nip33)
        if nip40:
            my_expire = ASQLiteNIP40Expirer(db=my_store.DB,
                                            interval=expire_interval)

    # Postgres store at the moment ios fucked and needs fixing... shouldn't be too much work
    elif store == 'postgres':
        my_store = get_postgres_store(db_name=args['pg_database'],
                                      user=args['pg_user'],
                                      password=args['pg_password'],
                                      is_nip16=nip16,
                                      is_nip33=nip33)
        # blank the password from printout
        args['pg_password'] = '***'
        if nip40:
            print('WARNING: NIP40 set true but currently no expirer for postgres so won\'t be enabled')

    elif store == 'transient':
        my_store = RelayMemoryEventStore(is_nip16=nip16,
                                         is_nip33=nip33)
        if nip40:
            my_expire = MemoryNIP40Expirer(interval=expire_interval,
                                           store=my_store)

    # if we have an event expirer start it
    if my_expire:
        asyncio.create_task(my_expire.run())
        relay_info['supported_nips'] = [40]


    # just running to empty db
    if is_wipe:
        if store not in ('transient', 'none'):
            my_store.destroy()
        else:
            print(f'{store} store, no action required!')
        sys.exit(0)

    # optional message accept handlers
    accept_handlers = []
    if nip42:
        accept_handlers.append(AuthenticatedAcceptor())
    if max_length:
        accept_handlers.append(LengthAcceptReqHandler(max=max_length))
    if max_before or max_after:
        accept_handlers.append(CreateAtAcceptor(max_before=max_before,
                                                max_after=max_after))


    for c_handler in accept_handlers:
        logging.info(c_handler)

    logging.debug(f'config = {args}')

    my_relay = Relay(my_store,
                     max_sub=max_sub,
                     accept_req_handler=accept_handlers,
                     ack_events=nip20,
                     relay_information=relay_info,
                     request_auth=nip42)

    ssl_context = None
    protocol = 'ws'
    if is_ssl:
        ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        ssl_context.load_cert_chain(ssl_cert, ssl_key)
        protocol = 'wss'

    def print_run_info():
        col_size = 40
        print(f'running relay at {protocol}://{host}:{port}{end_point} persiting events to store {store}')
        for k,v in my_relay.relay_information.items():
            print(f'{k.ljust(col_size)} {v}')
        # print(my_relay.relay_information)


    # access via tor?
    if enable_tor:
        with TORService(relay_port=port,
                        service_dir=tor_dir,
                        password=tor_password,
                        is_ssl=protocol=='wss',
                        empheral=tor_empheral) as my_tor:
            print_run_info()
            await my_relay.start(host=host,
                                 port=port,
                                 end_point=end_point,
                                 ssl_context=ssl_context)
    else:
        print_run_info()
        await my_relay.start(host=host,
                             port=port,
                             end_point=end_point,
                             ssl_context=ssl_context)


if __name__ == "__main__":
    logging.getLogger().setLevel(logging.ERROR)

    def sigint_handler(signal, frame):
        sys.exit(0)

    signal.signal(signal.SIGINT, sigint_handler)
    try:
        asyncio.run(main(get_args()))
    except ConfigError as ce:
        print(ce)




"""
usage: alias.py [-h] [-n] [-l] [-f FILENAME] [-k KEYS] profile_name

        link nostr keypairs to profile names

        alias.py <profile_name>           view existing mapping
        alias.py -n <profile_name>        map new keys auto generated
        alias.py -n <profile_name> <key>  map new with supplied key if pub_k then view only
        alias.py -l <profile_name> <key>  map existing to key, any exsting mapping overridden



positional arguments:
  profile_name          profile_name to perform action on

options:
  -h, --help            show this help message and exit
  -n, --new             create a new profile key pair link
  -l, --link            link key pair to exiting profile file, any existing mapping will be overridden
  -f FILENAME, --filename FILENAME
                        mappings in this file, default is {home}/profiles.csv
  -k KEYS, --keys KEYS  npub/nsec for the profile

"""
import logging
import asyncio
import sys
import argparse
from getpass import getpass
from pathlib import Path
from monstr.util import ConfigError
from monstr.ident.keystore import NamedKeys, SQLiteKeyStore, KeystoreInterface, KeyDataEncrypter
from monstr.encrypt import Keys, DecryptionException
from monstr_terminal.util import load_toml, get_sqlite_key_store

# defaults if not otherwise given
# working directory it'll be created it it doesn't exist
WORK_DIR = f'{Path.home()}/.nostrpy/'
# alias toml file
CONFIG_FILE = 'alias.toml'
# filename for key store
KEY_STORE_DB_FILE = 'keystore.db'

def print_k(k: NamedKeys, hex=True, bech32=True, include_private=True):
    print(k.name)
    if hex:
        print('** hex **')
        k.print_hex(include_private=include_private)
    if bech32:
        print('** bech32 **')
        k.print_bech32(include_private=include_private)


async def do_view(store: SQLiteKeyStore, name: str):
    k = await store.get(name)
    if k:
        print_k(k)
    else:
        print(f'profile not found: {name}')
    sys.exit(0)


async def do_delete(store: KeystoreInterface, name: str):
    if await store.get(name):
        yn = input(f'delete {name}, are you sure? ').lower()
        if yn == 'y':
            await store.delete(name)
            print(f'deleted {name}')
        else:
            print('not deleted')
    else:
        print(f'{name} not found to delete')
    sys.exit(0)


async def do_add(store: KeystoreInterface, name: str, keys: str):
    # new and generate keys
    if keys is None:
        keys = Keys()
    # new but keys supplied by user
    else:
        keys = Keys.get_key(keys)

    k = await store.add(k=keys,
                        name=name)

    print(f'created profile: {name}')
    print_k(k)


async def do_link(store: KeystoreInterface, name: str, keys: str):
    if keys is None:
        raise ConfigError('keys required to link')
    try:
        keys = Keys(keys)
        k = await store.update(k=keys,
                               name=name)

        print(f'linked keys to profile: {name}')
        print_k(k)
    except Exception as e:
        raise ConfigError(str(e))


async def do_list(store: KeystoreInterface):
    keys = await store.select()
    for k in keys:
        print(k)

def get_cmdline_args(args) -> dict:
    parser = argparse.ArgumentParser(description="""
            link nostr keypairs to profile names

            alias.py <profile_name>           view existing mapping
            alias.py -n <profile_name>        map new keys auto generated
            alias.py -n <profile_name> <key>  map new with supplied key if pub_k then view only
            alias.py -d <profile_name>        delete profile
            alias.py -l <profile_name> <key>  map existing to key, any exsting mapping overridden

        """, formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument('-n', '--new',
                        help='create a new profile key pair link',
                        action='store_true')
    parser.add_argument('-l', '--link',
                        help='link key pair to exiting profile file, any existing mapping will be overridden',
                        action='store_true')
    parser.add_argument('-d', '--delete',
                        help='delete an existing profile',
                        action='store_true')
    parser.add_argument('-a', '--all',
                        help='lists all existing profiles',
                        action='store_true')
    parser.add_argument('-f', '--filename',
                        default=args['keystore']['filename'],
                        type=str,
                        help=f'mappings in this file, default [{args["keystore"]["filename"]}]')
    parser.add_argument('profile_name',
                        nargs='?',
                        type=str,
                        help='profile_name to perform action on')
    parser.add_argument('-k',
                        '--keys',
                        help='npub/nsec for the profile',
                        default=None)
    parser.add_argument('--debug', action='store_true', help='enable debug output', default=args['debug'])
    return vars(parser.parse_args())


def get_args() -> dict:
    """
    get args to use order is
        default -> toml_file -> cmd_line options

    so command line option is given priority if given

    :return: {}
    """

    ret = {
        'keystore': {
            'filename': WORK_DIR + KEY_STORE_DB_FILE,
            'password': None
        },
        'debug': False
    }

    # now form config file if any
    load_toml(dir=WORK_DIR,
              filename=CONFIG_FILE,
              current_args=ret)

    # now from cmd line
    ret.update(get_cmdline_args(ret))

    # if debug flagged enable now and output args we're running with
    if ret['debug'] is True:
        logging.getLogger().setLevel(logging.DEBUG)
        logging.debug(f'get_args:: running with options - {ret}')

    return ret


async def profile_creator():
    args = get_args()
    profile_name = args['profile_name']
    keys = args['keys']

    # the key store, with encryptor for keys
    key_store = get_sqlite_key_store(db_file=args['keystore']['filename'],
                                     password=args['keystore']['password'])

    op_count = sum([args['delete'], args['new'], args['link']])
    if op_count > 1:
        print('only one of --new, --link or --delete can be set')
        sys.exit(1)

    try:
        # must be view
        if args['all'] is True:
            if profile_name is True:
                print('profile name ignored with view --all')
            await do_list(key_store)
        else:
            # get profile name as we need it
            if profile_name is None:
                profile_name = input('name: ')

            # if op count is 0 then must be view profile name
            if op_count == 0:
                if profile_name:
                    await do_view(store=key_store,
                                  name=profile_name)
            # else one of the ops must be true
            elif args['delete']:
                await do_delete(store=key_store,
                                name=profile_name)
            elif args['new']:
                await do_add(store=key_store,
                             name=profile_name,
                             keys=keys)
            elif args['link']:
                await do_link(store=key_store,
                              name=profile_name,
                              keys=keys)
    except ConfigError as ce:
        print(ce)
    except DecryptionException as de:
        print(f'bad password or non encrypted store? - {de}')
    except Exception as e:
        print(e)

# Press the green button in the gutter to run the script.
if __name__ == '__main__':
    logging.getLogger().setLevel(logging.ERROR)
    asyncio.run(profile_creator())








import asyncio
import logging
import os
import time

from abc import abstractmethod
from prompt_toolkit import Application
from prompt_toolkit.completion import WordCompleter
from prompt_toolkit.layout.containers import HSplit, Window, VSplit, WindowAlign
from prompt_toolkit.layout.controls import BufferControl, FormattedTextControl
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.layout.layout import Layout
from prompt_toolkit.layout.containers import FloatContainer, Float
from prompt_toolkit.layout import ScrollablePane
from prompt_toolkit.layout.containers import ConditionalContainer
from prompt_toolkit.widgets import VerticalLine, Button,TextArea, HorizontalLine, Dialog, \
    SearchToolbar, Frame, RadioList
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.filters import Condition
from prompt_toolkit.mouse_events import MouseEvent, MouseButton, MouseEventType
from monstr.ident.profile import Profile, UnknownProfile, ProfileList
from monstr.ident.event_handlers import ProfileEventHandler
from monstr.ident.persist import SQLProfileStore, MemoryProfileStore
from monstr.event.event import Event
from monstr.client.client import Client
from monstr.client.messaging import MessageThreads
from monstr.db.db import Database
from monstr.event.persist_memory import MemoryEventStore
from monstr.event.persist_sqlite import SQLEventStore

def is_left_click(e):
    return e.event_type == MouseEventType.MOUSE_DOWN and e.button == MouseButton.LEFT


class DialogBase:

    def __init__(self, title,
                 app: Application,
                 on_close=None):
        # screen layout obj, we'll get root con from here
        self._app = app
        # root con, it should be FloatContainer, maybe throw here if not...
        self._con = self._app.layout.container

        self._kb = KeyBindings()
        # probably at the least you'll want to focus somewhere
        self._on_close = on_close

        # key short cur to close the dialog
        @self._kb.add('c-c')
        def _(e):
            self.hide()

        self._title = title
        self._content = None
        self._my_diag = None
        self._my_float = None

    # the actual dialog obj that contains our content as it's body
    @abstractmethod
    def create_content(self):
        pass

    def show(self):
        self.create_content()
        # it seems these need to be recreated all the time
        self._my_diag = Dialog(title=self._title,
                               body=self._content)

        self._my_float = Float(self._my_diag)

        self._con.floats.append(self._my_float)

        self._app.layout.focus(self._my_diag)

    def hide(self):
        self._con.floats = []
        if self._on_close:
            self._on_close()


class SearchContactDialog(DialogBase):

    def __init__(self, app: Application,
                 profiles: ProfileList,
                 on_close=None):

        # inputs/widget
        # text area input contact search


        def my_cancel():
            self.hide()

        def my_select():
            if len(self._search_in.text) == 64:
                self._profile_key = self._search_in.text
            self.hide()

        self._search_in = None
        self._search_in_buf_con = None

        self._cancel_but = Button('cancel', handler=my_cancel)
        self._select_but = Button('select', handler=my_select)

        self._profiles = profiles
        self._highlight_index = 0
        super().__init__(title='start messaging contact',
                         app=app,
                         on_close=on_close)

        # couldn't work out how to get the autocomplete working in prompt toolkit in full screen app
        # so we'll do it ourself, maybe revisit in future

        self._match_con = None
        self._suggest_con = None
        self._matches = []
        self._search_text = ''

        # set when a valid key is entered and dialog closed with select
        self._profile_key = None

        @self._kb.add('down')
        def _(e):
            self._highlight_index += 1
            if self._highlight_index>=len(self._matches):
                self._highlight_index = 0
            self._prep_suggest_con()

        @self._kb.add('up')
        def _(e):
            self._highlight_index -= 1
            if self._highlight_index<0:
                self._highlight_index = len(self._matches)-1
            self._prep_suggest_con()


    def _prep_suggest_con(self):
        self._matches = []
        self._suggest_con = []
        in_str = self._search_text
        if len(in_str.replace(' ', '')) > 0:
            self._matches = self._profiles.matches(self._search_text, 20)

        to_add = []

        def get_complete(item_index):
            def my_complete(e):
                if is_left_click(e):
                    self._highlight_index = item_index
                    self._search_in.text = self._matches[self._highlight_index].public_key
                    self._app.layout.focus(self._select_but)
                    return True
            return my_complete

        for i, c_p in enumerate(self._matches):
            color = ''
            if i == self._highlight_index:
                color = 'green'

            to_add.append(Window(
                content=FormattedTextControl(text=[(color, c_p.display_name(with_pub=True), get_complete(i))]),
                height=1
            ))

        self._match_con.children = to_add

    def create_content(self):
        @Condition
        def _():
            return len(self._matches) > 0

        def my_change(buf: Buffer):
            self._highlight_index = 0
            self._search_text = buf.text
            self._prep_suggest_con()

        def my_complete(buf: Buffer):
            if self._matches:
                buf.text = self._matches[self._highlight_index].public_key
            self._app.layout.focus(self._select_but)
            return True

        self._search_in = Buffer(on_text_changed=my_change,
                                 multiline=False,
                                 accept_handler=my_complete)
        self._search_in_buf_con = BufferControl(self._search_in,
                                                preview_search=True)
        self._match_con = HSplit(children=[])
        self._suggest_con = ConditionalContainer(content=self._match_con, filter=_)

        self._content = HSplit(children=[
                Window(content=self._search_in_buf_con, width=32, height=1),
                self._suggest_con,
                Window(content=FormattedTextControl(text=''), width=32),
                VSplit(children=[
                    self._select_but,
                    Window(content=FormattedTextControl(text='')),
                    self._cancel_but
                ], height=1)
            ],
            key_bindings=self._kb)
        self._app.invalidate()

    def show(self):
        super().show()
        self._app.layout.focus(self._search_in)

    def hide(self):
        self._search_in.text = ''
        super().hide()

    @property
    def selected_profile_key(self):
        return self._profile_key


class SwitchProfileDialog(DialogBase):

    def __init__(self, app: Application,
                 on_profile_change,
                 on_close=None):
        self._on_profile_change = on_profile_change
        # inputs/widget
        # text area input contact search
        super().__init__(title='switch user',
                         app=app,
                         on_close=on_close)

    def create_content(self):
        to_add = []
        lookup = {}
        c_p: Profile
        for c_p in self._profiles:
            to_add.append(
                (c_p.public_key, c_p.display_name)
            )
            lookup[c_p.public_key] = c_p

        n_radio = RadioList(to_add, default=self._from_p.public_key)
        n_radio.show_scrollbar = False

        def is_changed():
            s_p = lookup[n_radio.current_value]
            if s_p != self._from_p:
                self._on_profile_change(s_p)
            self.hide()

        self._content = HSplit(children=[n_radio,
                                         VSplit(children=[
                                             Button(text='ok', handler=is_changed),
                                             Window(content=[], width=5),
                                             Button(text='cancel',handler=self.hide),
                                         ])],
                               key_bindings=self._kb)
        # self._content.children.append(TextArea())

    def show(self, from_p, all_profiles):
        self._from_p = from_p
        self._profiles = all_profiles
        super().show()


class RelayInfoDialog(DialogBase):

    def __init__(self, app: Application,
                 chat_app,
                 on_close=None):

        self._chat_app = chat_app
        self._relay_con = None
        super().__init__(title='relay status',
                         app=app,
                         on_close=on_close)

    def create_content(self):
        def do_close():
            self.hide()

        self._relay_con = HSplit([])
        self._content = HSplit([
            self._relay_con,
            VSplit([
                Window(),
                Button('ok', handler=do_close)
            ])
        ])
        self.update_relay_status()

    def update_relay_status(self):
        if self._content is None:
            return

        c_client: Client
        relay_info = []
        for c_client in self._chat_app.client:
            color = ''
            info_text = '%s - OK' % c_client.url
            if not c_client.connected:
                info_text = '%s - %s' % (c_client.url, c_client.last_error)
                color = 'red'

            relay_info.append(Window(FormattedTextControl(text=[(color, info_text)])))
        self._relay_con.children = relay_info

    def show(self):
        super().show()


class ChatGui:
    """
        interface for a simple 1 page app for viewing messages in the terminal
        using python prompt-toolkit
    """
    def __init__(self, chat_app):

        # app logic
        self._chat_app: ChatApp = chat_app

        self._msgs_height = 0

        # the root for everything, it's a float container so we can add dialogs
        self._root_con = FloatContainer(
            # main contents here
            content=Window(content=[]),
            # add dialogs here
            floats=[]
        )
        self._layout = Layout(self._root_con)

        kb = KeyBindings()

        @kb.add('c-q')
        def do_quit(e):
            self._app.exit()

        @kb.add('c-s')
        def do_post(e):
            my_send()

        @kb.add('c-up')
        def do_up(e):
            pos = self._scroll.vertical_scroll-1
            if pos < 0:
                pos = 0
            self._scroll.vertical_scroll = pos

        @kb.add('c-down')
        def do_down(e):
            pos = self._scroll.vertical_scroll

            if (self._msgs_height + 4 - pos) > os.get_terminal_size().lines:
                pos += 1

            self._scroll.vertical_scroll = pos

        @kb.add('tab')
        def do_tab(e):
            self._layout.focus_next()

        @kb.add('s-tab')
        def do_tab(e):
            self._layout.focus_previous()

        self._app = Application(layout=self._layout,
                                full_screen=True,
                                key_bindings=kb,
                                mouse_support=True)
        # create our dialogs
        # profile switch dialog
        self._p_switch_dialog = SwitchProfileDialog(self._app,
                                                    on_close=self._focus_prompt,
                                                    on_profile_change=self._chat_app.set_from_profile)
        # relay info dialog
        self._relay_info_dialog = RelayInfoDialog(self._app,
                                                  self._chat_app,
                                                  on_close=self._focus_prompt)

        # events
        def msg_entered(buffer):
            self._chat_app.do_message(buffer.text)
            buffer.text = ''
            return True

        def my_send():
            msg_entered(self._prompt.buffer)
            self._layout.focus(self._prompt)

        # self._prompt = Buffer(accept_handler=my_change,
        #                       multiline=True)  # Editable buffer.

        # msg text entered here

        self._prompt = TextArea(height=3,
                                accept_handler=msg_entered)

        self._msg_area = HSplit([])
        self._scroll = ScrollablePane(content=self._msg_area,
                                      keep_cursor_visible=True)

        # creates the struct for the left hand bar of screen
        self._nav_area = None
        self._nav_contacts = None
        self._nav_controls = None
        self._create_nav_pane()

        self._enter_bar = VSplit([
            # Window(height=3, content=BufferControl(buffer=self._prompt)),
            self._prompt,
            Button(text='send', handler=my_send)
        ])

        self._title = self._create_title()
        self._title['update_profile'](self._chat_app.profile)

        # now we have the parts ready actually construct the screen
        self._root_con.content = HSplit([
            self._title['title_con'],
            # content
            # ScrollablePane(self._main_window, keep_cursor_visible=True),
            # self._scroll,
            VSplit([
                # profile data here
                self._nav_pane,
                VerticalLine(),
                self._scroll
            ]),
            # msg entry
            self._enter_bar
        ])

        self._layout.focus(self._prompt)

    def _focus_prompt(self):
        self._layout.focus(self._prompt)

    def _create_title(self):
        profile_con = FormattedTextControl('')
        profile_win = Window(
            content=profile_con,
            align=WindowAlign.LEFT
        )
        status_con = FormattedTextControl('<connection status>')
        status_win = Window(
            content=status_con,
            align=WindowAlign.RIGHT
        )

        con_con = VSplit(children=[
            profile_win,
            status_win
        ], height=1)

        def switch_profile(e):
            if is_left_click(e):

                self._p_switch_dialog.show(from_p=self._chat_app.profile,
                                           all_profiles=self._chat_app.get_local_profiles())

        def relay_click(e):
            if is_left_click(e):
                self._relay_info_dialog.show()

        def update_profile(from_p: Profile):
            profile_text = '<no profile>'
            if from_p:
                profile_text = self._chat_app.profile.display_name(with_pub=True)

            child_arr = [
                ('', 'Nostrpy CLI message v0.1, user: '),
                ('green', profile_text, switch_profile)
            ]
            profile_con.text = child_arr

        def update_status():
            status_text = [('red', 'not connected!', relay_click)]
            if self._chat_app.connected:
                con_counts = self._chat_app.connect_count
                if con_counts[0] == con_counts[1]:
                    status_text = [('green', 'connected', relay_click)]
                else:
                    status_text = [('orange', 'connected %s/%s' % (con_counts[1], con_counts[0]), relay_click)]

            status_con.text = status_text

        return {
            'title_con': con_con,
            'update_profile': update_profile,
            'update_status': update_status
        }

    def update_title(self):
        # we could expose this seperately but doesn't seem worth it
        self._title['update_profile'](self._chat_app.profile)
        self._title['update_status']()
        self._relay_info_dialog.update_relay_status()
        self._app.invalidate()

    def _create_nav_pane(self):
        def new_contact():
            self._new_contact_dialog.show()

        def contact_selected():
            if self._new_contact_dialog.selected_profile_key:
                self._chat_app.set_view_profile(self._new_contact_dialog.selected_profile_key)


            self._focus_prompt()

        self._new_contact_dialog = SearchContactDialog(self._app,
                                                       profiles=self._chat_app.get_profile_lookup(),
                                                       on_close=contact_selected)

        self._nav_contacts = HSplit([])
        self._nav_controls = HSplit([
            Button(text='new contact', handler=new_contact),
            # Button(text='relays'),
            # Button(text='switch user')
        ])

        self._nav_pane = HSplit([
            self._nav_controls,
            HorizontalLine(),
            ScrollablePane(content=self._nav_contacts),

        ], width=24)

    def set_contacts(self, contacts):
        to_add = []

        def get_click(p: Profile):
            def my_click(e: MouseEvent):
                if is_left_click(e) and self._chat_app.set_view_profile:
                    self._chat_app.set_view_profile(p)

            return my_click

        for c in contacts:
            c_p = c['profile']
            n_count = ''
            if c['new_count']:
                n_count = '(%s)' % c['new_count']

            c_text = '%s %s' % (c_p.display_name(), n_count)
            color = ''
            if self._chat_app.view_profile and c_p.public_key == self._chat_app.view_profile.public_key:
                color = 'green'

            to_add.append(Window(content=FormattedTextControl(text=
            [
                (color, c_text, get_click(c_p))
            ])
                , height=1)
            )
        # just added to extend to the bottom of the screen
        to_add.append(Window())
        self._nav_contacts.children = to_add
        self._app.invalidate()

    def run(self):
        self._app.run()

    def _get_msg_prompt(self, msg: Event, to: Profile):
        to_display = to.display_name()
        from_p = self._chat_app.profile
        from_display = from_p.display_name()

        prompt_width = max(
            len(from_display),
            len(to_display),
            10
        )

        messager = from_p
        if msg.pub_key != from_p.public_key:
            messager = to

        u_dispay = messager.display_name()

        if len(u_dispay) > prompt_width:
            u_dispay = u_dispay[:u_dispay - 2] + '..'

        prompt_text = '%s@%s:' % (u_dispay.rjust(prompt_width),
                                  msg.created_at)

        prompt_col = 'gray'
        if messager != self._chat_app.profile:
            prompt_col = 'green'

        return prompt_col, prompt_text

    def update_messages(self):
        """
            redraws the message for chat_apps current view (to) profile
        """
        self._msg_area.children = []
        to_p = self._chat_app.view_profile
        c_msg: Event
        if not to_p:
            self._msgs_height = 1
            self._scroll.vertical_scroll = 0
            self._msg_area.children.append(
                Window(content=FormattedTextControl(text='click on a user or create new contact'), height=1)
            )
        else:
            msgs = self._chat_app.messages
            # height calc based on \n line count, going to be a bit flaky, there must be a better way?!
            total_height = 0
            for c_msg in msgs:
                c_msg_arr = []
                prompt_col, prompt_text = self._get_msg_prompt(c_msg, to_p)
                c_msg_arr.append((prompt_col, prompt_text))

                first_line = True
                for c_line in c_msg.content.split('\n'):
                    if first_line:
                        c_msg_arr.append(('', c_line))
                        first_line = False
                    else:
                        c_msg_arr.append(('', '\n' + ''.join([' ']*len(prompt_text)) + c_line))

                # c_msg_arr.append(('', '\n'))
                # msg_arr.append('[SetCursorPosition]', '')
                win_height = len(c_msg_arr)-1
                n_win = Window(content=FormattedTextControl(text=c_msg_arr), height=win_height)
                total_height += win_height
                self._msg_area.children.append(n_win)

            self._msgs_height = total_height
            self._app.invalidate()

            if self._msgs_height + 4 <= os.get_terminal_size().lines:
                self._scroll.vertical_scroll = 0
            else:
                self._scroll.vertical_scroll = self._msgs_height - os.get_terminal_size().lines + 4


class ChatApp:

    def __init__(self, as_profile,
                 client,
                 db: Database,
                 kind=Event.KIND_ENCRYPT):

        # either client or client pool (untested..)
        self._client = client

        self._db = db
        # helps us present friendly ident info rather then having to show keys
        self._profiles: ProfileEventHandler = None

        if self._db:
            # we'll want to listen eventually and add as handler to client sub
            self._profiles = ProfileEventHandler(SQLProfileStore(self._db), None)
            self._event_store = SQLEventStore(self._db)
        else:
            self._profiles = ProfileEventHandler(MemoryProfileStore(), None)
            self._event_store = MemoryEventStore()

        self._profile = self._profiles.profiles.get_profile(as_profile,None)

        # nothing generate either using passed in value as priv_key or if that isn't valid generate one
        if not self._profile:
            if as_profile is not None:
                self._profile = self._profiles.profiles.get_profile(as_profile,
                                                                create_type='private')
            if not self._profile:
                adhoc_keys = Profile.get_new_key_pair()
                self._profile = Profile(priv_k=adhoc_keys['priv_k'],
                                        profile_name='adhoc_profile')
            # needed to show in switch profile
            self._profiles.profiles.put(self._profile)

        if not self._profile or not self._profile.private_key:
            raise UnknownProfile('unable to find profile %s or we don\'t have the private key for it')
        logging.debug('starting chat app using profile: %s' % self._profile)

        # standard txt or encrypted, to add wrapped inbox
        self._kind = kind

        # init gui
        self._display = ChatGui(chat_app=self)

        # in here we keep message threads and contacts per profile
        self._threads = {}
        # we do the same thing for contacts per profile
        self._contacts = {}

        # self._current_to = self._get_profile('3648e5c206883d9118d9c19a01ddde96059c5f46a89444b252e247ca9b9270e3')
        self._current_to = None
        # move to connect...
        self._draw_contacts()
        self._draw_messages()

        self._status = {
            'connected': None,
            'relay_count': 0,
            'connect_count': 0
        }

        self._start_client()

    def _start_client(self):
        # at the moment we're passing in actual client obj, think better to pass in urls and create ourself

        def my_connect(the_client: Client):
            handlers = [self]

            # we always create version of this now they just might not be to perm store
            handlers.append(PersistEventHandler(self._event_store))
            handlers.append(self._profiles)

            # because we added switching profiles simpler just to look at all events of correct kind
            filter = {
                'kinds': [self._kind, Event.KIND_META]
            }
            # to stop us always asking for everything, in the case of using a pool this might result in
            # missing events from certain relays so maybe add as option to Client where it'll get add since as
            # newest events per relay before making the subscribe. To do this clients will need to take the eventstore
            # on init, we could then also then just make some of the persistance stuff as bool flags on create?
            if self._event_store:
                evts = self._event_store.get_filter(filter)
                filter['since'] = self._event_store.get_newest(the_client.url)
                for c_evt in evts:
                    self.do_event(None, c_evt, None)

            the_client.subscribe(handlers=handlers, filters=filter)


        def my_status(status):
            last_connect_count = self.connect_count
            self._status = status
            if last_connect_count != self.connect_count:
                self._display.update_title()


        self._client.set_on_connect(my_connect)
        self._client.set_status_listener(my_status)
        self._client.start()

    @property
    def connected(self):
        return self._status['connected']

    @property
    def connect_count(self):
        return self._status['relay_count'], self._status['connect_count']

    @property
    def client(self):
        return self._client

    def do_event(self, sub_id, evt: Event, relay):
        # we only need to forward evts for profiles we already looked at, as others
        # will be picked up on load from db or on create msgthread
        to_p = evt.p_tags + [evt.pub_key]
        for p_k in to_p:
            if p_k in self._threads:
                self._threads[p_k].do_event(sub_id, evt, relay)
                break

    def do_message(self, text):
        self.profile_threads.post_message(the_client=self._client,
                                          from_user=self._profile,
                                          to_user=self._current_to,
                                          text=text,
                                          kind=self._kind)

    def _draw_contacts(self):
        profiles = []
        for c_key in self.profile_contacts:
            contact = self.profile_contacts[c_key]
            profiles.append({
                'profile': self._get_profile(c_key,
                                             create_type='public'),
                'new_count': len(self.profile_threads.messages(pub_k=c_key,
                                                               kind=self._kind,
                                                               since=contact['last_view'],
                                                               received_only=True))
            })

        self._display.set_contacts(profiles)

    def _new_message(self, msg: Event):
        if msg.pub_key not in self.profile_contacts \
                or msg.p_tags[0] not in self.profile_contacts:
            pub_k = msg.pub_key
            if pub_k == self.profile.public_key:
                pub_k = msg.p_tags[0]

            self.add_contact(pub_k)

        # new message on profile we're looking at
        if self._current_to and msg.pub_key == self._current_to.public_key or msg.pub_key == self._profile.public_key:
            self._draw_messages()
        else:
            self._draw_contacts()

    def _draw_messages(self):
        if self._current_to:
            msgs = self.profile_threads.messages(self._current_to.public_key,
                                                 self._kind)
            if msgs:
                self.profile_contacts[self._current_to.public_key]['last_view'] = msgs[len(msgs) - 1].created_at

        self._display.update_messages()

    def set_view_profile(self, p):
        """
        :param p: key or profile obj
        :return:
        """
        if isinstance(p, str):
            p = self._get_profile(p, create_type='public')

        # nothing to do if we're already msging and setting view to profile we're using not allowed
        if p != self._current_to and p != self._profile:
            # add to contacts if not already there
            self.add_contact(p.public_key)
            self._current_to = p
            self._draw_messages()
            self._draw_contacts()

    def set_from_profile(self, p: Profile):
        self._profile = p
        self._contacts = {}

        # to back to nothing
        self._current_to = None

        # update display for this new profile
        self._draw_messages()
        self._draw_contacts()
        self._display.update_title()

    @property
    def profile(self) -> Profile:
        # profile we're currently logged in as, the from profile
        return self._profile

    @property
    def profile_threads(self) -> MessageThreads:
        # threads for the currently logged in profile
        if self.profile.public_key not in self._threads:
            # creates new thread if we don't already have it for this profile
            self._threads[self.profile.public_key] = MessageThreads(from_p=self._profile,
                                                                    evt_store=self._event_store,
                                                                    on_message=self._new_message,
                                                                    kinds=self._kind)

        return self._threads[self.profile.public_key]

    @property
    def profile_contacts(self):
        if self.profile.public_key not in self._contacts:
            to_add = self._contacts[self.profile.public_key] = {}
            # make the intial list by looking at what messages we've received
            for c_key in self.profile_threads.messaged():
                if c_key not in self._contacts:
                    to_add[c_key] = {
                        'last_view': None
                    }

        return self._contacts[self.profile.public_key]

    def add_contact(self, pub_k):
        # add contact for the profile we're currently using
        if pub_k not in self.profile_contacts:
            self.profile_contacts[pub_k] = {
                'last_view': None
            }

    @property
    def view_profile(self) -> Profile:
        # msgs we're viewing and where any post will be sent, the to profile
        return self._current_to

    def get_profile_lookup(self) -> ProfileList:
        # obj that gives us a way to look up profiles as we currently know them
        return self._profiles.profiles

    @property
    def messages(self):
        # this will return msgs between current from/to profiles
        # at the moment we recreate threads on switching profile
        return self.profile_threads.messages(self._current_to.public_key,
                                             kind=self._kind)

    def get_local_profiles(self):
        # profiles where we have a private key, i.e. we can make posts
        ret = [pp for pp in self._profiles.profiles if pp.private_key]
        return ret

    def _get_profile(self, as_profile, create_type=None):
        # if str maybe its a profile_name, privkey or pubkey
        ret = None

        # we were handed a profile obj so everything is probably cool...
        if isinstance(as_profile, Profile):
            ret = as_profile
        # ok assuming we have a db lets see if we can find this profile
        elif isinstance(as_profile, str) and self._profiles:
            ret = self._profiles.profiles.lookup_priv_key(as_profile)
            if not ret:
                ret = self._profiles.profiles.lookup_profilename(as_profile)
            if not ret:
                ret = self._profiles.profiles.lookup_pub_key(as_profile)

        # we didn't find a profile but we'll see if we can just use as priv key...
        # also fallback we don't have db
        if not ret and create_type is not None:
            if len(as_profile) == 64:
                # and also hex, will throw otherwise
                bytearray.fromhex(as_profile)
                if create_type == 'private':
                    ret = Profile(priv_k=as_profile,
                                  profile_name='adhoc_user')
                elif create_type == 'public':
                    ret = Profile(pub_k=as_profile)

        return ret

    def start(self):
        try:
            self._display.run()
        except Exception as e:
            logging.debug(e)




from __future__ import annotations

import logging
import asyncio
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from monstr.client.client import Client

from monstr.client.event_handlers import DeduplicateAcceptor, NotOnlyNumbersAcceptor
from monstr.event.event import Event
from monstr.ident.profile import Profile
from monstr.ident.event_handlers import ProfileEventHandlerInterface
from monstr.inbox import Inbox
from monstr.signing.signing import SignerInterface
from monstr.encrypt import Keys


class PostApp:

    def __init__(self,
                 use_relay,
                 profile_handler: ProfileEventHandlerInterface,
                 user_sign: SignerInterface,
                 to_users_k: [Keys],
                 inbox: Inbox=None,
                 subject=None,
                 is_encrypt=True,
                 kind=None,
                 tags=None
                 ):
        """
        :param as_user:     posts made as this user
        :param to_users:    posts will be made to these users, required if doing encrypted posts but can be set to None
                            when is_encrypt is False - in this case you'll just be in broadcast sending posts out
        :param inbox:       send events wrapped through this inbox account
        :param subject:     subject tag will be added to msgs and used as a filter also to see replies
        :param is_encrypt:  if true then NIP4 encoded

        TODO: we should probably do some checking on init values, at the moment we're expecting th ecaller to do

        """
        self._client = use_relay
        self._profile_handler = profile_handler
        self._user_sign = user_sign
        self._to_users_k = to_users_k
        self._inbox = inbox
        self._subject = subject

        self._is_encrypt = is_encrypt

        # default kinds for encrypt/no encrypt
        if kind is None:
            if self._is_encrypt:
                self._kind = Event.KIND_ENCRYPT
            else:
                self._kind = Event.KIND_TEXT_NOTE

        # not, doesn't stop you from encrypting data send e.g. over kind 1 though you probably don't want to do that
        else:
            self._kind = kind

        # tags to be added to every post
        self._tags = tags

        # de-duplicating of events for when we're connected to multiple relays
        self._acceptors = [
            DeduplicateAcceptor(),
            # bit crap needs the message to be exacly the same, better if it was using matching patterns and with whitelist
            # DuplicateContentAcceptor(),
            # if content is just numbers then reject, again whitelist would be good
            NotOnlyNumbersAcceptor()
        ]

        # all the mesages we've seen, if since and event store then may be some from before we started
        self._msg_events = []
        self._on_msg = None

        # this are init'd by the _get_ready_task
        self._as_user: Profile = None
        self._as_user_pub_k = None
        self._to_users_p = []
        self._chat_members = []

        asyncio.create_task(self._get_ready())
        self._ready = False

    async def wait_ready(self, max_wait: int = 3):
        waited = 0.0
        while self._ready is False and int(waited) < max_wait:
            await asyncio.sleep(0.1)
            waited += 0.1

        if self._ready is False:
            raise Exception('PostApp:: wait_ready, something went wrong during setup...')

    async def _create_chat_members(self):
        chat_members = set([await self._user_sign.get_public_key()])

        k: Keys
        if self._to_users_k:
            chat_members = chat_members.union(set([k.public_key_hex() for k in self._to_users_k]))

        chat_members = list(chat_members)
        chat_members.sort()
        self._chat_members = chat_members

    async def _get_ready(self):
        # our pk
        self._as_user_pub_k = await self._user_sign.get_public_key()
        # pks we're messageing
        to_ks = []
        # ks for profiles we're going to fetch
        p_to_fetch = [self._as_user_pub_k]

        if self._to_users_k:
            to_ks = [k.public_key_hex() for k in self._to_users_k]
            p_to_fetch += to_ks

        if self._inbox:
            p_to_fetch = p_to_fetch + [await self._inbox.pub_key]

        # pre-fetch any profiles we'll need, create stubbs if they can't be found
        await self._profile_handler.aget_profiles(pub_ks=p_to_fetch,
                                                  create_missing=True)

        # get a profile obj for sending user (us)
        self._as_user = self._profile_handler.get_profile(pub_k=self._as_user_pub_k)

        # get profile only for anyone we're sending to
        self._to_users_p = self._profile_handler.get_profiles(pub_ks=to_ks)

        # create chat members, used to create the view filter
        await self._create_chat_members()

        self._ready = True

        # evts are put into a async queue to be dealt with
        self._msg_queue = asyncio.Queue()
        asyncio.create_task(self._my_consumer())

    async def _my_consumer(self):
        try:
            while True:
                args = await self._msg_queue.get()

                await self.do_event_task(*args)
        except Exception as e:
            logging.debug(f'PostApp::_my_consumer - {e}')

    def _is_chat(self, msg: Event):
        """
        is this msg part of the chat we're looking at, currently this is just made
        up by have all the correct members in it, so if all the members are the same
        then you're looking in that group...
        TODO: look at proper group chat NIP and implement

        :param msg:
        :return:
        """

        msg_members = list(set([msg.pub_key]).union(msg.p_tags))
        msg_members.sort()

        if self._subject:
            is_subject = False
            if msg.get_tags('subject'):
                is_subject = self._subject in [s[0] for s in msg.get_tags('subject')]
        else:
            is_subject = True

        return self._chat_members == msg_members and is_subject or (self._is_encrypt is False and not self._to_users_k)

    def accept_event(self,
                     the_client: Client,
                     sub_id: str,
                     evt: Event) -> bool:
        ret = True
        for c_accept in self._acceptors:
            try:
                ret = c_accept.accept_event(the_client, sub_id, evt)
                if not ret:
                    break
            except Exception as e:
                logging.debug(e)
        return ret

    def do_event(self, client: Client, sub_id, evt: Event):
        self._msg_queue.put_nowait((client, sub_id, evt))

    async def do_event_task(self, client: Client, sub_id, evt: Event):
        if self._inbox:
            try:
                # if plaintext we won't send a singer in which will stop us unwrapping encrypted
                # events event if we could
                user_sign = None
                if self._is_encrypt:
                    user_sign = self._user_sign
                # return None if we failed to unwrap
                evt = await self._inbox.unwrap_event(evt, user_sign)
            except Exception as e:
                pass

        if evt and self.accept_event(client, sub_id, evt) and evt.kind == self._kind:
            if self._is_chat(evt):
                self._msg_events.append(evt)
                if self._on_msg:
                    self._on_msg(evt)

    def set_on_message(self, callback):
        self._on_msg = callback

    async def do_post(self, msg):
        for evt in await self.make_post(msg):
            self._client.publish(evt)

    async def show_post_info(self, msg: str=None):

        if msg is None:
            msg = '<no msg supplied>'
        just = 10
        print('from:'.rjust(just), self.as_user.display_name())
        if self._to_users_p:
            p: Profile
            print('to:'.rjust(just), [p.display_name() for p in self._to_users_p])
        if self._inbox:
            print('via:'.rjust(just), await self._inbox.name)

        if self._subject:
            print('subject:'.rjust(just), self._subject)

        enc_text = 'encrypted'
        if not self._is_encrypt:
            enc_text = 'plain_text'
        print('format:'.rjust(just), enc_text)
        if self._kind not in {1, 4}:
            print('kind :'.rjust(just), self._kind)

        print('%s\n%s\n%s' % (''.join(['-'] * 10),
                              msg,
                              ''.join(['-'] * 10)))

    async def make_post(self, msg) -> [Event]:
        """
        makes post events, a single event if plaintext or 1 per to_user if encrypted
        :param public_inbox:
        :param as_user:
        :param msg:
        :param to_users:
        :param is_encrypt:
        :param subject:
        :return:
        """
        async def make_event(post_to: str = None):
            ret = Event(kind=self._kind,
                        content=msg,
                        pub_key=self._as_user_pub_k,
                        tags=tags)

            if self._is_encrypt:
                ret.content = await self._user_sign.nip4_encrypt(plain_text=ret.content,
                                                                 to_pub_k=post_to)

            await self._user_sign.sign_event(ret)

            if self._inbox:
                ret = await self._inbox.wrap_event(ret,
                                                   from_sign=self._user_sign,
                                                   to_k=post_to)

            return ret

        tags = []
        if self._to_users_p:
            tags = [['p', p.public_key] for p in self._to_users_p]

        if self._tags:
            tags = tags + self._tags

        if self._subject is not None:
            tags.append(['subject', self._subject])

        try:
            if not self._is_encrypt:
                post = [await make_event()]

            # when it's encrypt we make as many event as to_users
            else:
                post = []
                for c_post in tags:
                    if c_post[0] == 'subject':
                        continue
                    # we leave all the p_tags - should we just restrict to who we're sending too?
                    post.append(await make_event(c_post[1]))


        except Exception as e:
            print(e)

        return post

    @property
    def message_events(self):
        return self._msg_events

    @property
    def as_user(self) -> Profile:
        return self._as_user

    @property
    def as_signer(self) -> SignerInterface:
        return self._user_sign

    @property
    def connection_status(self):
        return self._client.connected




"""
usage: profile_search.py [-h] [-r RELAY] [-a AS_USER] [-b BOOTSTRAP] [-s SINCE] [-d]

search for nostr user profiles

options:
  -h, --help            show this help message and exit
  -r RELAY, --relay RELAY
                        comma separated urls of relays to connect to - default ws://localhost:8081
  -a AS_USER, --as AS_USER
                        nsec/npub/hex or alias for account viewing as - default None
  -b BOOTSTRAP, --bootstrap BOOTSTRAP
                        nsec/npub/hex or alias for accounts used for bootstrapping - default None
  -s SINCE, --since SINCE
                        n days to search back for profile events - default 5
  -d, --debug           enable debug output


"""
import logging
import asyncio
import sys
import aioconsole
import argparse
from datetime import datetime, timedelta
from monstr.util import util_funcs
from monstr.ident.persist import MemoryProfileStore
from monstr.ident.event_handlers import NetworkedProfileEventHandler
from monstr.ident.profile import Profile, Contact
from monstr.ident.alias import ProfileFileAlias
from monstr_terminal.cmd_line.util import FormattedEventPrinter
from monstr.client.event_handlers import EventHandler
from monstr.client.client import ClientPool, Client
from monstr.event.event import Event
from monstr.encrypt import Keys
from monstr.exception import ConfigurationError
from pathlib import Path

# defaults if not otherwise given
# working directory it'll be created it it doesn't exist
WORK_DIR = '%s/.nostrpy/' % Path.home()
# relay/s to attach to
DEFAULT_RELAY = 'ws://localhost:8081'
# profiles persited here sqlite db
DB = WORK_DIR + 'monstr.db'
# nmae > key alias here
ALIAS_FILE = WORK_DIR + 'profiles.csv'
# how far to fetch metas back too in days
DEFAULT_SINCE = 5


def get_args():
    parser = argparse.ArgumentParser(
        prog='profile_search.py',
        description='search for nostr user profiles'
    )
    parser.add_argument('-r', '--relay', action='store', default=DEFAULT_RELAY,
                        help='comma separated urls of relays to connect to - default %s' % DEFAULT_RELAY)
    parser.add_argument('-a', '--as_user', action='store', default=None,dest='as_user',
                        help='nsec/npub/hex or alias for account viewing as - default %s' % None)
    parser.add_argument('-b', '--bootstrap', action='store', default=None,
                        help='nsec/npub/hex or alias for accounts used for bootstrapping - default %s' % None)
    parser.add_argument('-s', '--since', action='store', default=90, type=int,
                        help='n days to search back for profile events - default %s' % DEFAULT_SINCE)
    parser.add_argument('-d', '--debug', action='store_true', help='enable debug output')

    ret = parser.parse_args()

    if ret.debug:
        logging.getLogger().setLevel(logging.DEBUG)

    my_alias = ProfileFileAlias(ALIAS_FILE)
    p: Profile = None
    if ret.as_user:
        k = Keys.get_key(ret.as_user)
        if k is None:
            p = my_alias.get_profile(ret.as_user)
            if p is None:
                raise ConfigurationError('as_user - %s not nostr key and no matching alias' % ret.as_user)
            k = p.keys
        ret.as_user = k

    if ret.bootstrap:
        boot_keys = []
        for c_k in ret.bootstrap:
            k = Keys.get_key(c_k)
            if k is None:
                if p is None:
                    raise ConfigurationError('bootstrap - %s not nostr key and no matching alias' % c_k)
                k = p.keys
            boot_keys.append(k)
        ret.bootstrap = boot_keys

    return ret


async def display_profile(p: Profile,
                          profile_handler,
                          as_user,
                          output: str='short',
                          key_output='npub'):
    follows = ''
    follow_folow_count = ''
    c_contact: Contact
    f_p: Profile

    if as_user:
        if not as_user.contacts_is_set():
            await profile_handler.aload_contacts(as_user)
        # for marking those we follow
        if p.public_key in as_user.contacts.follow_keys():
            follows = 'F'

        count_ff = 0

        await profile_handler.get_profiles(pub_ks=as_user.contacts.follow_keys(),
                                           create_missing=True)
        for c_k in as_user.contacts.follow_keys():
            f_p = await profile_handler.get_profile(c_k)
            if f_p.contacts_is_set():
                if p.public_key in set(f_p.contacts.follow_keys()):
                    count_ff += 1

        follow_folow_count = 'FF%s' % count_ff

    pub_key = p.keys.public_key_bech32()
    if key_output != 'npub':
        pub_key = p.keys.public_key_hex()

    if output == 'short':
        await aioconsole.aprint(p.display_name(False)[0:15].rjust(15), ' - ', pub_key, ' ', follows, ' ', follow_folow_count)
    elif output == 'long':
        margin = 10
        await aioconsole.aprint('%s - %s' % ('name'.ljust(margin),
                                             p.display_name()))
        await aioconsole.aprint('%s - %s' % ('npub'.ljust(margin),
                                             p.keys.public_key_bech32()))
        await aioconsole.aprint('%s - %s' % ('hex'.ljust(margin),
                                             p.public_key))
        if p.get_attr('about'):
            await aioconsole.aprint('%s - %s' % ('about'.ljust(margin),
                                                 p.get_attr('about')))
        if p.get_attr('picture'):
            await aioconsole.aprint('%s - %s' % ('picture'.ljust(margin),
                                                 p.get_attr('picture')))
        if p.get_attr('banner'):
            await aioconsole.aprint('%s - %s' % ('banner'.ljust(margin),
                                                 p.get_attr('banner')))

        for c_attr in p.attrs:
            if c_attr not in ('picture', 'banner', 'name', 'about'):
                await aioconsole.aprint('%s - %s' % (('%s' % c_attr).ljust(margin),
                                                     p.get_attr(c_attr)))

        await aioconsole.aprint('%s - %s' % ('link'.ljust(margin),
                                             'https://hamstr.to/profile/%s' % p.keys.public_key_bech32()))


    else:
        await aioconsole.aprint(p.as_dict())


class My_EventHandler(EventHandler):

    def __init__(self,
                 profile_handler,
                 since):

        self._profile_handler = profile_handler
        self._since = since

        self._u_pubs_client = {}
        self._my_task = None

        super().__init__()

    @property
    def since(self):
        return self._since

    @since.setter
    def since(self, val):
        self._since = val

    def do_event(self, client: Client, sub_id, evt: [Event]):
        if isinstance(evt, Event):
            evt = [evt]

        if client.url not in self._u_pubs_client:
            self._u_pubs_client[client.url] = set()

        u_pubs = self._u_pubs_client[client.url]

        # if meta update now
        self._profile_handler.do_event(client, sub_id, Event.latest_events_only(evt))

        # for all other events extract pub_k and p_tag
        c_evt: Event
        other_events = [c_evt for c_evt in evt if c_evt.kind != Event.KIND_META]

        for c_evt in other_events:
            u_pubs.add(c_evt.pub_key)
            for pub_k in c_evt.p_tags:
                u_pubs.add(pub_k)

        async def do_fetches():
            await asyncio.sleep(0.2)

            to_fetch = u_pubs
            self._u_pubs_client[client.url] = set()

            for c_chunk in util_funcs.chunk(list(to_fetch), 20):
                await self._profile_handler.aget_profiles(c_chunk)
            self._since = util_funcs.date_as_ticks(datetime.now())
            self._my_task = None

        # now start a job to fetch those we don't have
        if self._my_task is None:
            self._my_task = asyncio.create_task(do_fetches())


async def init_user(user:Keys, profile_handler) -> Profile:
    ret: Profile = await profile_handler.get_profile(user.public_key_hex(),
                                                    create_missing=True)
    await aioconsole.aprint('running as - %s' % ret.display_name())
    await profile_handler.load_contacts(ret)

    # init contacts for those we follow too
    asyncio.create_task(profile_handler._fetch_contacts(ret.contacts.follow_keys()))

    return ret


async def do_search():
    try:
        args = get_args()
    except ConfigurationError as ce:
        print(ce)
        sys.exit(2)

    relay = args.relay.split(',')
    profile_store = MemoryProfileStore()
    # we'll look at events since here and onwards as we run to get pub_ks
    # you probably don't want to do many as it could be a lot of events
    since = util_funcs.date_as_ticks(datetime.now() - timedelta(minutes=args.since))
    as_user: Profile = None
    sub_id = None

    def do_sub():
        nonlocal sub_id
        filters = [{
            'kinds': [Event.KIND_META, Event.KIND_CONTACT_LIST,
                      Event.KIND_TEXT_NOTE, Event.KIND_CHANNEL_MESSAGE],
            'since': my_handler.since
        }]

        sub_id = my_client.subscribe(
            sub_id=sub_id,
            handlers=[
                my_handler
            ],
            filters=filters)

    def on_connect(the_client: Client):
        print('connect!!!??!?!?!?')
        do_sub()

    # def my_eose(the_client: Client, sub_id: str, events: [Event]):
    #     # add eose because it means these well get done in batch
    #     my_handler.do_event(the_client, sub_id, events)
    #     print('initial events loaded...')

    async def do_command(cmd: str):
        cmd_split = cmd.split()
        cmd = cmd_split[0]
        args = cmd_split[1:]
        output_formats = ('long', 'json', 'short')

        async def _show_profile():
            if output not in output_formats:
                await aioconsole.aprint('invalid output format %s - valid values %s' % (output,
                                                                                        output_formats))
                return

            if args:
                to_show = []
                for c_a in args[0].split(','):
                    k = Keys.get_key(c_a)
                    if k is None:
                        await aioconsole.aprint('%s is not a valid nostr key' % c_a)
                    else:
                        to_show.append(k)
                p: Profile
                for c_k in to_show:
                    p = await peh.get_profile(c_k.public_key_hex())
                    if p:
                        await display_profile(p,
                                              output=output,
                                              profile_handler=peh,
                                              as_user=as_user)
                    else:
                        await aioconsole.aprint('profile not found for key - %s' % c_k.public_key_bech32())

            else:
                await aioconsole.aprint('no keys supplied')

        if cmd == 'count':
            await aioconsole.aprint('%s profiles in store' % len(profile_store.select_profiles()))
        elif cmd == 'profile':
            output = 'long'
            if len(args) > 1:
                output = args[1]
            await _show_profile()
        elif cmd == 'contacts':
            output = 'short'
            for_user = as_user
            if len(args) == 1:
                k = Keys.get_key(args[0])
                if k:
                    for_user = await peh.aget_profile(k.public_key_hex())
                else:
                    output = args[0]
            elif len(args) == 2:
                k = Keys.get_key(args[0])
                if k:
                    for_user = await peh.aget_profile(k.public_key_hex())
                output = args[1]
            if for_user:
                if not for_user.contacts_is_set():
                    await peh.aload_contacts(for_user)
                c_contact: Contact
                await peh.aget_profiles([c_contact.contact_public_key for c_contact in for_user.contacts],
                                        create_missing=True)
                for c_contact in for_user.contacts:
                    c_p = await peh.aget_profile(c_contact.contact_public_key)
                    await display_profile(c_p,
                                          profile_handler=peh,
                                          as_user=as_user,
                                          output=output)
            else:
                await aioconsole.aprint('unable to get for_user for contacts')
        elif cmd == 'posts':
            for_user = as_user
            if args:
                k = Keys.get_key(args[0])
                if k:
                    for_user = await peh.aget_profile(k.public_key_hex())
            if for_user:
                events = await my_client.query({
                    'authors': [for_user.public_key],
                    'kinds': [Event.KIND_TEXT_NOTE],
                    'limit': 10
                })
                c_evt: Event
                for c_evt in events:
                    await post_printer.aprint_event(the_client=None,
                                                    sub_id=None,
                                                    evt=c_evt)
                    # await aioconsole.aprint('%s@%s' % (util_funcs.str_tails(c_evt.id), c_evt.created_at))
                    # await aioconsole.aprint(c_evt.content)
            else:
                await aioconsole.aprint('unable to get for_user for post')

        else:
            await aioconsole.aprint('unknown command - %s' % cmd)

    # start the client
    my_client = ClientPool(relay)
    asyncio.create_task(my_client.run())
    await my_client.wait_connect(timeout=10)
    peh = NetworkedProfileEventHandler(client=my_client,
                                       store=profile_store)
    my_handler = My_EventHandler(
        profile_handler=peh,
        since=since
    )

    # do as_user and boot user lookups before subscribing
    boot_pubs = []
    if args.as_user:
        boot_pubs.append(args.as_user.public_key_hex())
    if args.bootstrap:
        boot_pubs = boot_pubs + [k.public_key_hex() for k in args.bookeys]

    # actually fetch
    await my_client.query({
        'kinds': [Event.KIND_CONTACT_LIST, Event.KIND_META],
        'authors': boot_pubs
        },
        do_event=my_handler.do_event
    )

    if args.as_user:
        as_user = await init_user(args.as_user,
                                  peh)

    # used by $posts
    post_printer = FormattedEventPrinter(profile_handler=peh,
                                         as_user=as_user)
    # now add
    my_client.set_on_connect(on_connect)
    do_sub()

    msg = ''
    c_m: Profile
    while msg != 'exit':
        msg = await aioconsole.ainput('> ')
        msg = msg.lower()
        if msg.replace(' ','') == 'exit':
            continue
        elif msg and msg[0] == '$':
            await do_command(msg[1:])
        else:
            search_filter = {}
            if msg:
                search_filter['name'] = msg
                if msg.startswith('npub') or msg.startswith('nsec'):
                    try:
                        search_filter['public_key'] = Keys.get_key(msg).public_key_hex()
                    except Exception as e:
                        logging.debug(f'bad pr partial key {msg}?')
                else:
                    search_filter['public_key'] = msg
                search_filter['about'] = msg

            matches = profile_store.select_profiles(search_filter)
            if matches:
                for c_m in matches:
                    await display_profile(c_m,
                                          profile_handler=peh,
                                          as_user=as_user)
            else:
                await aioconsole.aprint('no matches found!')


if __name__ == "__main__":
    util_funcs.create_work_dir(WORK_DIR)
    util_funcs.create_sqlite_store(DB)
    asyncio.run(do_search())

